%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%  File:        lapopt.sl
%
%  Description: Code Optimization
%               Reformation of jump and load sequences
%
%  Author:      H. Melenk, ZIB Berlin
%  Date:        22 Dec 1986
%
% Copyright (c) 1993, Konrad Zuse Zentrum Berlin, all rights reserved
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% 

(fluid '(*trlapopt &function-has-quotes&))

(setq *trlapopt nil)

%---------------------------------------------------------------------
%  cmacro optimizer
%---------------------------------------------------------------------

(de lapopt1 (u)
    (prog()
        (setq &function-has-quotes& t)
	(when (not *lapopt) (return u))
        (look-for-quotes)
        (setq u (lapopt-move-special-cases u))
	(return u)))

%
% determine whether the function uses base registers
%  (quote, internalentry)
%


(de look-for-quotes(u)
   (setq &function-has-quotes& nil)
   (look-for-quotes1 u))

(de look-for-quotes1(w)
   (cond(&function-has-quotes& nil)
        ((atom w)
         (when (or (eq w 'quote) 
                   (InternallyCallableP w)
                   (eq w NAME&)       % auto-recursive call needs base register
               )
               (setq &function-has-quotes& t))
       )
       ((eq (car w) '!*entry) nil)    % dont parse (*entry ...) statement
       (t (look-for-quotes1 (car w))
          (look-for-quotes1 (cdr w)) )))


%
%  Move bodies of if-then-elseif ... sequences  such that at runtime as few
%  jumps as possible are taken.
%

(fluid '(&cond-cm&))

(setq &cond-cm& 
 '((*jumpeq      . *jumpnoteq)       (*jumpnoteq      . *jumpeq)
   (*jumpwgeq    . *jumpwlessp)      (*jumpwlessp     . *jumpwgeq)
   (*jumpwleq    . *jumpwgreaterp)   (*jumpwgreaterp  . *jumpwleq)
   (*jumptype    . *jumpnottype)     (*jumpnottype    . *jumptype)
   (*jumpintype  . *jumpnotintype)   (*jumpnotintypte . *jumpintype)
   (*jumpeqtag   . *jumpnoteqtag)    (*jumpnoteqtag   . *jumpeqtag)
   (*jumpwgeqtag . *jumpwlessptag)   (*jumpwlessptag  . *jumpwgeqtag)
  % no inverse jumps for
   (*jumpwgreaterptag)
   (*jumpon)
   
))

(de lapopt-move-special-cases(code)
  (let (u w lbl lab rcode bcode inst)
   (while (not (lapopt-bottom code))
    (setq inst (pop code))
     (when 
      (and
       (pairp inst)
       (setq u (atsoc (car inst) &cond-cm&))
       (setq u (cdr u))
       (setq lbl (cadr inst))
       (or (memq lbl code)(member (setq lbl (list '*lbl lbl)) code))
       (setq w (lapopt-move-special-cases1 code lbl))
      )
      (setq lab (list 'label (gensym)))
      (setq bcode (nconc bcode (cons (list '*lbl lab) (car w))))
      (push (cons u (cons lab (cddr inst))) rcode)
      (setq inst nil)
      (setq code (cdr w))
     )
     (when inst (push inst rcode))
    )
  (setq code (nconc (reversip rcode) (nconc bcode code)))
  (when *trlapopt 
         (prin2t "=== move special cases:")
         (MAPCAR code 'PRINT))
  code
)) 

(de lapopt-bottom(u)
 (or (null u) 
     (eqcar (car u) 'wordalign)))
 
(de lapopt-move-special-cases1 (code lbl)
  % Find basic block until lbl which ends by an unconditional jump.
  % Return nil or pair of bblock and rest of code.
  (let (rcode fcode inst op)
   (while code
     (setq inst (pop code))
     (push inst rcode)
     (cond ((or (atom inst) (equal inst lbl)) (setq code nil))
           ((and (memq (setq op (car inst)) 
                      '(*jump *linke *exit))
                 (cdr code)
                 (equal (car code) lbl)
            )
            (setq fcode code) (setq code nil))
           ((or (eq op '*lbl) (atsoc op &cond-cm&))
            (setq code nil)
           )
   ))
   (when fcode (cons (reversip rcode) fcode))
))


%---------------------------------------------------------------------
%  instruction optimizer
%---------------------------------------------------------------------

%
% Code align: align function entry point to octaword
%


(de code-align(u)
  (prog (w a i f)
    (setq w u)
    (setq a (if *writingfaslfile currentoffset* codebase*))
    (setq f '(fullword 0))
 loop
    (when (null u)(return w))
    (setq i (car u))
    (cond ((atom i)(go cont))
          ((eq (car i) 'fullword)
           (setq f i) (setq a (wplus2 a 8)) (go cont))
          ((eq (car i) '*entry) (go entry))
    )
 cont
    (setq u (cdr u))
    (go loop)
 entry
    (while (not (weq 0 (wand a 16#f)))
           (setq a (wplus2 a 8))
           (setf (cdr u)(cons (car u) (cdr u)))
           (setf (car u) f))
    (return w) 
))

% operation:
%
% remove jumps to jumps
% remove nop from delay slot
% try to reform load operations

(compiletime (progn
	(on fast-integers)
	(flag '(operands LapOptRemoveToteHosen lookForLabelForwards
	 lookForLabelBackwards label-from  read-label) 'internalfunction)))
 
(fluid '(*ConditionalJumps* *unconditionalJumps* *goodjumps* % *badjumps*
	 *nonMovables* *nonMovables1* *alljumps* *lapopt *lapopttrace))
 
(setq *goodjumps* '( Br bri call calli bnc!.t bc!.t))

(setq *conditionalJumps* '((Bc!.t . Bnc!.t) (Bc!.t . Bnc!.t)) )
 
(setq *unconditionalJumps* '(Br bri call calli))

(setq *nonMovables* (append *goodjumps!* '(
 MOVE-RELOCATABLE-OBJECT MOVE-RELOCATABLE-OBJECT-and-load
 MOVE-RELOCATABLE-OBJECT-and-store nop *entry call )))

(setq *nonMovables1* (append *goodjumps!* '( nop *entry call )))

(setq *alljumps* (append *goodjumps!* *unconditionalJumps*))

(compiletime (ds NOPp (x) ((lambda(h) (and (pairp h)(eq (car h) 'NOP))) x)))

(de LapOptReformCalls (code)
  % 1. move
  %   (LDQ (REG R27) (DISPLACEMENT (REG SYMFNCCENTER) (XIDLOC PRINT)))
  %    followed by
  %   (JMP ...)  or (JSR ...)
  %   as far backwards as possible
  %
  %  2. eliminate restauration of base register after call when not needed.
  %
  (let (rcode inst inst2 lastinst hasquote)
     (while code
       (setq inst (pop code))
       (cond
         ((not (pairp inst)) t)
         ((null rcode) t) 
           % case 1.
         ((and 
            (pairp (setq lastinst (car rcode)))
            (eqcar inst 'LDQ)
            (equal (cadr inst) '(REG R27))
            (eqcar (caddr inst) 'DISPLACEMENT)
            (eqcar (caddr (caddr inst)) 'XIDLOC) )
          (setq rcode (LapOptReformCalls1 rcode inst))
          (setq inst nil)
         )
           % case 2.
         ((and
            (equal inst ' 
              (LDA (REG R27) (DISPLACEMENT (REG RETURNADDR) NEARESTENTRY*)))
            (stringp (car code))
            (or 
              (not &function-has-quotes&)
              (not (needs-base-register (cdr code)))
            )
          )
          (setq inst nil)
         )
       )
       (when inst (setq rcode (cons inst rcode)))
     )
     (reversip rcode) ))

(de needs-base-register(w)
   (cond ((null w) nil)
         ((atom (car w)) t)
         ((branchp (caar w)) t)
         ((eq (caar w) 'JSR) nil)
         ((and (eq (caar w) 'LDQ) (equal (cadar w) '(REG R27))) nil)
         ((and (eq (caar w) 'S8ADDQ) (equal (cadddr (car w)) '(REG R27))) nil)
         ((critical-instp(car w) '(R27 QUOTE INTERNALENTRY)) t)
         (t (needs-base-register (cdr w)))
))

(de branchp(c)
  (let ((x (get c '*format*)))
   (and (pairp x)
        (memq (car x)
          '(*memory_format_branch* *branch_format*)))))
   
(de critical-instp(u l)
  (if (atom u) 
      (or (stringp u) (memq u l))
      (or (critical-instp (car u) l)
          (critical-instp (cdr u) l) )))

(de LapOptReformCalls1 (rcode inst)
  % push inst down into rcode as far as possible
   (cond ((null rcode)(list inst))
         ((or (atom (car rcode))
              (memq (caar rcode) '(*LBL* *ENTRY))
              (critical-instp (car rcode) 
                  '(R27 QUOTE INTERNALENTRY))
              (branchp (caar rcode))
          )
          (cons inst rcode)
         )
         (t (cons (car rcode) 
                  (LapOptReformCalls1 (cdr rcode) inst))) ))
              
%===========================================================

(de LapOptReformJumps (code)
  (let (rcode rrcode inst inst2 x y z cond jump inv lab lastinst delayslot)
       (return code) % Main switch

%        % first phase% change (any (without reg))
%                              (ld (reg)
%                              (any (using reg))
% into (ld ..) (any ..) (any ..)

      (while code
	 (setq inst (pop code))
	 (when (and (pairp inst) rcode (pairp (setq lastinst (car rcode)))
		(eqcar lastinst 'ld!.l) 
		(member (caddr lastinst) (operands inst))
		(cdr rcode) (pairp (setq inst2 (cadr rcode)))
		(not (member (caddr lastinst) (operands inst2)))
		(or (atom (cddr rcode)) (atom (caddr rcode))
		    (not (member (caaddr rcode) *alljumps*)))
		% dont "steal" from delay slot
		(or (atom (cdr inst2)) 
		    (not (member (lastcar inst2) (operands lastinst))))
		% target of first instr must not be needed for second
		(not (and (eqcar inst2 'st!.l)
			  (equal (caddr inst2) (cadr lastinst))))
		% no intermix of ld!.l and st!.l of same object
		(not (member (car inst2) *nonMovables1*)))
		(setq x (pop rcode)) (setq y (pop rcode))
		(push x rcode) (push y rcode))
	(push inst rcode))

      (setq code (reversip rcode)) (setq rcode nil)

	% second phase: look for jumps to jumps and for negation jumps
     (setq delayslot 0)
     (while code
	  (setq inst (pop code)) 
	  (when (and (pairp inst)    % negation jump?
		(setq x (atsoc (car inst) *conditionalJumps*)) 
		(setq lab (label-from  inst))
		 code (NOPp (car code) )
		(pairp (setq y (cdr code))) 
		(pairp (setq inst2 (car y))) (eq (car inst2) 'Br)
		(cdr y) (NOPp  (cadr y))
		(setq z(cddr y)) (equal (car z) lab))
	    (setq inst (cons (cdr x) (cdr inst2)))
	    (setq code (cons '(NOP) z))     )
	  (while(and (pairp inst)    % jump to jump ?
		(memq (car inst) *goodJumps*)
		(setq lab (label-from inst))  
		code (equal (car code) '(NOP))
		(setq x (or (lookForLabelForwards code lab)
			    (lookForLabelBackwards rcode lab)))
		(eqcar (car x) 'Br)
		(NOPp (cadr x) )
	     (setq inst (cons (car inst) (cdar x)))    )  )
	  (push inst rcode))  
     (setq code (reversip rcode)) (setq rcode (setq rrcode (cons nil nil))) 
	  % third phase: fill in delayed instructions
     (while code
	 (setq inst (pop code)) 
	 (setq delayslot (cond ((eq delayslot 3) 2)
			       ((eq delayslot 2) 1)
			       ((eq delayslot 1) 0)
			       (t 0)))
	 (while(and (pairp inst)    % jump to jump ?
		(memq (car inst) *goodJumps*)
		(setq lab (label-from inst))  
		code (NOPp  (car code) ) 
		(setq x (or (lookForLabelForwards code lab)
			    (lookForLabelForwards rcode lab)))
		(setq inst2 (car x))
		(pairp inst2)  
		(not (memq (car inst2) *nonMovables*)))
				     % we insert a label
	    (when(not(atom(setq lab (cadr x))))
		 (setq lab (gensym)) (rplacd x (cons lab (cdr x))))
	    (rplacd rrcode  (setq z `((,(car inst) ,lab))))
     % used to be 3
	    (setq delayslot 2) %(setq lastinst (car rrcode))
	    (setq rrcode z)
	    (setq inst (car x))
	    (setq code (cdr code)))
	(when % last case
	      % (something) (BR lbl)(NOP)
	      % -> (BR lbl) (something)
	      % if (something) is no label and is movable
	    (and (pairp inst) (memq (car inst) *unconditionalJumps*)
		 code 
		 (NOPp (car code))
		      % if not 2nd successor of a jump
		 %or (null lastinst)
		 %   (and (pairp lastinst)
		 %        (not (memq (car lastinst) *allJumps*))))
		 (eq 0 delayslot)
		 (not (and (memq (car inst) '(call calli))
			   (member '(reg RETURNADDR) (operands (car  rrcode)))))
		 (not (member (cadr inst) (operands (car  rrcode))))
		 (pairp (car rrcode)) 
		 (not(memq (caar rrcode) *nonMovables*)))
	    (setq code (cdr code))
	    (setq delayslot 3) % (setq lastinst inst)
	    (setq y (car rrcode)) (rplaca rrcode inst)(setq inst y)
	    )
	(rplacd rrcode  (setq z (cons inst nil)))
	(when (and (pairp inst) (memq (car inst) *alljumps*))
	       (setq delayslot 3)) % this instruction has a delay slot
	(setq rrcode z))
	   % done
	(LapOptRemoveToteHosen (cdr rcode))
  ))

% since the jumps are (BA hugo) and not (ba (label hugo)
% and it does not handle the jumps backward correctly (see tak example
(de LapOptRemoveToteHosen (code)
   (prog (l labels rcode inst prevjump1 prevjump2 alllabs)
       (setq alllabs (mapcan code
		   (function (lambda(inst) (when (atom inst) (list inst))))))
       (mapc code (function (lambda(inst)
	    (when (pairp inst)
	     (setq l (read-label inst alllabs))
	     (when (not(member l labels)) (setq labels (cons l labels)))))))
       (while code
	      (setq inst (pop code))
	      (while (and (atom inst)                   % is a label
			  (not (member inst labels))    % no reference
			  prevjump2)                    % previous is a jump
		     (while (and code (pairp (setq inst (pop code))))))
	      (setq prevjump2 prevjump1)
	      (setq prevjump1 (when (pairp inst)
				    (memq (car inst) '(Br bri))))
	      (push inst rcode))
       (return(reversip rcode)) ))

(de lookForLabelForwards(code lab)
       (cond ((null code) nil)
	     ((equal (car code) lab)(cdr code))
	     (T (lookForLabelForwards (cdr code) lab))))
 
(de lookForLabelBackwards(code lab)
       (cond ((null code) nil) 
	     ((null (cdr code)) nil)
	     ((null (cddr code)) nil)
	     ((equal (caddr code) lab)(list (cadr code) (car code)))
	     (T (lookForLabelBackwards (cdr code) lab))))
	 
(de label-from (inst) 
     (setq inst (cadr inst))
     (if (atom inst) inst (cadr inst)))
     
(de read-label (u alllabs)
    (cond ((atom u) (if (member u alllabs) u nil))
	  ((read-label (car u) alllabs))
	  (T (read-label (cdr u) alllabs))))

(de operands (x)         %simplified form , only looking for regs
   % extract the operands from x
   (cond ((atom x) nil)
	 ((member x '((reg zero) (reg 0))) nil)
	 ((or (memq (car x) '(reg))) (list x))
	 ((memq (car x) '(displacement label wconst indexed deferred))
	  (operands (cdr x)))
	 (t (nconc (operands (car x)) (operands (cdr x))))))

(de lapoptdec (u)
  (if (not *lapopt) u
  (let (inst new_inst (bb nil) (in 0) cdb (codebase* codebase*)
	 (bittableoffset* bittableoffset*)  rest_u
	 curo (currentoffset* currentoffset*) forwardinternalreferences*)
  (setq nearestentry* 0)
  (setq  cdb codebase* curo currentoffset*)
  (while (and u (not (eqcar inst '*entry))) % strip header until entry
	(setq inst (pop u))
        (push inst new_inst))
  (setq  new_inst (reverse new_inst))
  (while u
	(setq inst (pop u)) 
        (cond	((or (atom inst)	 % label; end of basic block
		     (eqcar inst '*entry))
			(setq new_inst 
			   (append new_inst
				(lapo (wplus2 cdb curo) in (reverse bb))))
			(setq in 0 bb nil)
			(setq codebase* cdb currentoffset* curo)
			(setq new_inst (append  new_inst (list inst))))
		((eq in 290)    % block is full
			(setq u (cons inst u))
			(setq new_inst
                           (append new_inst
                                (lapo (wplus2 cdb curo) in (reverse bb))))
                        (setq in 0 bb nil)
                        (setq codebase* cdb currentoffset* curo))
		((eqcar inst 'WORDALIGN) (setq rest_u u) (setq u nil))
		(t (depositinstruction inst) (setq bb (cons inst bb))
					     (setq in (iadd1 in))))
  )
  (setq new_inst (append new_inst (lapo (wplus2 cdb curo) in (reverse bb))))
  (append new_inst (cons (list 'WORDALIGN) rest_u)))))

(compiletime (put 'halfword_getmem 'opencode
                '((ldl (reg 1) (displacement (reg 1) 0))))
             (put 'halfword_putmem 'opencode
                '((stl (reg 2) (displacement (reg 1) 0)))))

(de lapo (adr in bb)
 (if (wlessp (length bb) 2) bb
  (prog (index newlist)
   (echoon adr in adr (if *lapopttrace 10 1))
   (setq in (isub1 in))
   (for (from i 0 in 1)
	(do (setq index (halfword_getmem (wplus2 (wtimes2 4 i) adr)))
	    (setq newlist (cons (nth bb (iadd1 index)) newlist))))
  (return (reverse newlist)))))
