%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         $pxc/spectrum-instrs.sl
% Description:  PSL Instructions for Spectrum (HP RISC) 
% Author:       Winfried Neun
% Created:      19 September 1991
% Mode:         Lisp
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1991, Konrad Zuse Zentrum Berlin, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Revisions
%
%  25. Jan 1993 (Winfried Neun)
% Added check of long conditional jumps
%

% Creating the format functions
% 36 Formats are a lot of them!

(dm define-format (uu)

   (prog (u number item size sum arglist putfields)

       (setq u (cdr uu))
       (setq number (pop u))
       (setq sum 0)
       (while u
          (setq item (pop u))
          (setq size  (pop u))
          (unless (numberp size) (errorinformat))
          (if (numberp item)
	   (setq putfields (cons `(setf (field word ,sum ,size) ,item)
				putfields))
           (setq putfields (cons `(setf (field word ,sum ,size)
				         (getinfo ',item arglist))
				putfields)))
          (setq sum (plus2 sum size))
       )

       (unless (eq sum 32) (errorinformat))
       (return `(de ,(intern (bldmsg "%w-%d" 'formatfn number)) (nam arglist)
	             (prog (word)
				   (setq word 0)
                         ,@ putfields
				   (depositword word))))))


(de getinfo (item arglist)
  (prog (val)
   (setq val (assoc item arglist))
   (when (null val) (print (list "no info " item arglist))
		    (error 99 "kaputt"))
   (setq val (cdr val))
   (return
     (cond ((regp val) (get-register-code val))
            (t val)))))

% Format 1: Load and Stores

(define-format 1 op 6 b 5 t!/r 5 s 2 im14 14)

% Format 2: Indexed Loads

(define-format 2 op 6 b 5 x 5 s 2 u 1 0 1 0 2 ext4 4 m 1 t 5)

% Format 3: Short displacement loads

(define-format 3 op 6 b 5 im5 5 s 2 a 1 1 1 0 2 ext4 4 m 1 t 5)

% Format 4: Short displacement stores, Store bytes Short

(define-format 4 op 6 b 5 r 5 s 2 a 1 1 1 0 2 ext4 4 m 1 im5 5)

% Format 5: Long immediates

(define-format 5 op 6 t!/r 5 im21 21)

% Format 6: Arithmetic/Logical

(define-format 6 op 6 r2 5 r1 5 c 3 f 1 ext7 7 t 5)

% Format 7: Arithmetic immediate

(define-format 7 op 6 r 5 t 5 c 3 f 1 e 1 im11 11)

% Format 8: Extract

(define-format 8 op 6 r 5 t 5 c 3 ext3 3 p 5 clen 5)

% Format 9: Deposit

(define-format 9 op 6 t 5 r!/im5 5 c 3 ext3 3 cp 5 clen 5)

% Format 10: Shift

(define-format 10 op 6 r2 5 r1 5 c 3 ext3 3 cp 5 t 5)

% Format 11: Conditional Branch

(define-format 11 op 6 r2!/p 5 r1!/im5 5 c 3 w1 11 n 1 w 1)

% Format 12: Branch External Branch and Link External

(define-format 12 op 6 b 5 w1 5 sr 3 w2 11 n 1 w 1)

% Format 13: Branch and Link, Gateway

(define-format 13 op 6 t 5 w1 5 ext3 3 w2 11 n 1 w 1)

% Format 14: Branch and Link Register, Branch vectored

(define-format 14 op 6 t!/b 5 x 5 ext3 3 0 11 n 1 0 1)

% Format 15: Data memory Management, Probe

(define-format 15 op 6 b 5 r!/x!/im5 5 s 2 ext8 8 m 1 t 5)

% Format 16: Instruction Memory Management

(define-format 16 op 6 b 5 r!/x!/im5 5 sr 3 ext7 7 m 1 0 5)

% Format 17: Break

(define-format 17 op 6 im13 13 ext8 8 im5 5)

% Format 18: Diagnose

(define-format 18 op 6 im26 26)

% Format 19: Move to/from Space Register

(define-format 19 op 6 0 5 r!/0 5 sr 3 ext8 8 0!/t 5)

% Format 20: Load Space ID

(define-format 20 op 6 b 5 0 5 s 2 0 1 ext8 8 t 5)

% Format 21: Move to Control register

(define-format 21 op 6 t 5 r 5 0 3 ext8 8 0 5)

% Format 22: Move from Control Register

(define-format 22 op 6 r 5 0 5 0 3 ext8 8 t 5)

% Format 23: System Control

(define-format 23 op 6 0!/b 5 0!/r!/im5 5 0 3 ext8 8 0!/t 5)

% Format 24: Special Operation Zero

(define-format 24 op 6 sop1 15 0 2 sfu 3 n 1 sop2 5)

% Format 25: Special Operation One

(define-format 25 op 6 sop1 15 0 2 sfu 3 n 1 t 5)

% Format 26: Special Operation Two

(define-format 26 op 6 r 5 sop1 10 2 2 sfu 3 n 1 sop2 5)

% Format 27: Special Operation Three

(define-format 27 op 6 r2 5 r1 5 sop1 5 3 2 sfu 3 n 1 sop2 5)

% Format 28: Coprocessor Operation

(define-format 28 op 6 sop1 17 uid 3 n 1 sop2 5)

% Format 29: Coprocessor Indexed Loads

(define-format 29 op 6 b 5 x 5 s 2 u 1 0 1 0 2 0 1 uid 3 m 1 t 5)

% Format 30: Coprocessor Indexed Stores

(define-format 30 op 6 b 5 x 5 s 2 u 1 0 1 0 2 1 1 uid 3 m 1 r 5)

% Format 31: Coprocessor Short Displacement Loads

(define-format 31 op 6 b 5 im5 5 s 2 a 1 1 1 0 2 0 1 uid 3 m 1 t 5)

% Format 32: Coprocessor Short Displacement Stores

(define-format 32 op 6 b 5 im5 5 s 2 a 1 1 1 0 2 1 1 uid 3 m 1 r 5)

% Format 33: Floating Point Operation Zero

(define-format 33 op 6 r 5 0 5 sop 3 fmt 2 0 2 0 3 0 1 t 5)

% Format 34: Floating Point Operation One

(define-format 34 op 6 r 5 0 4 sop 2 df 2 sf 2 1 2 0 3 0 1 t 5)

% Format 35: Floating Point Operation Two

(define-format 35 op 6 r1 5 r2 5 sop 3 fmt 2 2 2 0 3 n 1 c 5)

% Format 36: Floating Point Operation Three

(define-format 36 op 6 r1 5 r2 5 sop 3 fmt 2 3 2 0 3 0 1 t 5)

% Format 37: "Linked triad" operations

(define-format 37 op 6 rm1 5 rm2 5 ta 5 ra 5 f 1 tm 5)

%------------------------------------------------

(dm def-in (u)
   (prog (name suffixes item operands suffix-list)
	(setq suffix-list '( cd1!|cd2 cd1 cd2 cd3 cd4 cd5 cd6
			     sfu sop uid n cp1 cp2 cp3 cp4 fmt sf df ))
	(setq u (cdr u))
        (setq name (pop u))
        (while (memq (setq item (pop u)) suffix-list)
		(push item suffixes)
		)
	(setq operands (cons item nil))

        (while (not (numberp (setq item (pop u))))
		(push item operands))
		
        (return `(progn 
		   (put ',name 'suffixes ',(reverse suffixes))
		   (put ',name '*op-length* 4)
		   (put ',name 'operands ',(reverse operands))
		   (put ',name 'format ,item )
                   (put ',name 'opcode ,(pop u))
		   ,(if u `(put ',name 'extension ,(car u))
			(prog2 
		 	 (when (memq item
			      '(2 3 4 6 7 8 9 10 13 14 15 16 17 19 20 21 22 23))
			    (printf " %w needs an extention code %n" name))
					nil)
)))))

%       Name  completers  operands             format opcode extension(opt)

(def-in	add	cd2	r1	r2	t	6	02	16#30)
(def-in	addbf	cd2	n	r1	r2	tgt	11	16#2a)
(def-in	addbt	cd2	n	r1	r2	tgt	11	16#28)
(def-in	addc	cd2	r1	r2	t	6	02	16#38)
(def-in	addco	cd2	r1	r2	t	6	02	16#78)
(def-in	addi	cd2	im11	r	t		7	16#2d 16#0)
(def-in	addibf	cd2	im5	r2	tgt	11	16#2b)
(def-in	addibt	cd2	n	im5	r2	tgt	11	16#29)
(def-in	addil		im21	r	5	16#0a)
(def-in	addio	cd2	im11	r	t		7	16#2d 16#1)
(def-in	addit	cd2	im11	r	t		7	16#2c 16#0)
(def-in	addito	cd2	im11	r	t		7	16#2c 16#1)
(def-in	addl	cd2	r1	r2	t	6	02	16#50)
(def-in	addo	cd2	r1	r2	t	6	02	16#70)
(def-in	and	cd3	r1	r2	t	6	02	16#10)
(def-in	andcm	cd3	r1	r2	t	6	02	16#00)
	
(def-in	bb	cd5	n	r1	p	tgt	11	16#31)
(def-in	be	n	wd (sr	b)	12	16#38)
(def-in	bl	n	target	t	13	16#3a	16#0)
(def-in	ble	n	wd (sr	b)	12	16#39)
(def-in	blr	n	x	t	14	16#3a	16#2)
(def-in	break	i1	i2	17	0	16#0)
(def-in	bv	n	x (b)	14	16#3a	16#6)
(def-in	bvb	cd5	n	r1	tgt	11	16#30)
	
(def-in	comb	cd1	n	r1	r2	tgt	11	16#22)
(def-in	combf	cd1	n	r1	r2	tgt	11	16#22)
(def-in	combt	cd1	n	r1	r2	tgt	11	16#20)
(def-in	comclr	cd1	r1	r2	t	6	16#2	16#44)
(def-in	comib	cd1	n	im5	r2	tgt	11	16#23)
(def-in	comibf	cd1	n	im5	r2	tgt	11	16#23)
(def-in	comibt	cd1	n	im5	r2	tgt	11	16#21)
(def-in	comiclr	cd1	i	r	t	7	16#24	16#44)
	
(def-in	dcor	cd4	r	t	6	16#2	16#5c)
(def-in	dep	cd5	r	p	len	t	9	16#35	16#3)
(def-in	depi	cd5	im5	p	len	t	9	16#35	16#7)
(def-in	diag	i	18	16#05)
(def-in	ds	cd1!|cd2	r1	r2	t	6	16#02	16#22)
	
(def-in	extrs	cd5	r	p	len	t	8	16#34	16#7)
(def-in	extru	cd5	r	p	len	t	8	16#34	16#6)
	
(def-in	fdc	cp4	x (s b)	15	01	16#4a)
(def-in	fdce	cp4	x (s b)	15	01	16#4b)
(def-in	fic	cp4	x (sr	b)	16	01	16#a)
(def-in	fice	cp4	x (sr	b)	16	01	16#b)

(def-in	gate	n	target	t	13	16#3a	16#1)

(def-in	idcor	cd4	r	t	6	02	16#5e)
(def-in	idtlba	r	(s b)	15	01	16#41)
(def-in	idtlbp	r	(s b)	15	01	16#40)
(def-in	iitlba	r	(sr b)	16	01	16#1)
(def-in	iitlbp	r	(sr b)	16	01	16#0)

(def-in	ldb	d	(s b)	t	1	16#10)
(def-in	ldbs	cp2	d	(s b)	t	3	03	16#0)
(def-in	ldbx	cp1	x	(s b)	t	2	03	16#0)
(def-in	ldcws	cp2	d	(s b)	t	3	03	16#07)
(def-in	ldcwx	cp1	x	(s b)	t	2	03	16#7)
(def-in	ldh	d (s b)	t	1	16#11)
(def-in	ldhs	cp2	d (s b)	t	3	03	16#1)
(def-in	ldhx	cp1	x (s b)	t	2	03	16#1)
(def-in	ldil	im21	t	5	08)
(def-in	ldo	d(b)	t	1	16#0d)
(def-in	ldsid	(s b)	t	20	00	16#85)
(def-in	ldw	d	(s b)	t	1	16#12)
(def-in	ldwas	cp2	d	(b)	t	3	03	6)
(def-in	ldwax	x(b)	t	2	03	16#6)
(def-in	ldwm	d	(s b)	t	1	16#13)
(def-in	ldws	cp2	d	(s b)	t	3	03	16#2)
(def-in	ldwx	cp1	x	(s b)	t	2	03	16#2)
(def-in	lha	cp4	x(s b)	t	15	01	16#4c)
(def-in	lpa	cp4	x	(s b)	t	15	01	16#4d)


(def-in mfctl	r	t	22	00	16#45)
(def-in mfsp	sr	t	19	00	16#25)
(def-in movb	cd5	n	r1	r2	tgt	11	16#32)
(def-in movib	cd5	n	i	r2	tgt	11	16#33)
(def-in mtctl	r	t	21	00	16#c2)
(def-in mtsm	r	23	00	16#c3)
(def-in mtsp	r	sr	19	00	16#c1)
 
(def-in or	cd3	r1	r2	t	6	02	16#12)
 
(def-in pdc	cp4	x (s b)	15	01	16#4e)
(def-in pdtlb	cp4	x (s b)	15	01	16#48)
(def-in pdtlbe  cp4	x (sr b)	15	01	16#49)
(def-in pitlb	cp4	x (sr b)	16	01	16#8)
(def-in pitlbe  cp4	x (sr b)	16	01	16#9)
(def-in prober	(s b)	r	t	15	01	16#46)
(def-in proberi	(s b)	i	t	15	01	16#c6)
(def-in probew	(s b)	r	t	15	01	16#47)
(def-in probewi	(s b)	i	t	15	01	16#c7)
 
(def-in rfi	NIL	17	00	16#60)
(def-in rsm	(i t)	23	00	16#73)
 
(def-in sh1add cd2	r1	r2	t	6	02	16#32)
(def-in sh1addl cd2	r1	r2	t	6	02	16#52)
(def-in sh1addo cd2	r1	r2	t	6	02	16#72)
(def-in sh2add cd2	r1	r2	t	6	02	16#34)
(def-in sh2addl cd2	r1	r2	t	6	02	16#54)
(def-in sh2addo cd2	r1	r2	t	6	02	16#74)
(def-in sh3add cd2	r1	r2	t	6	02	16#36)
(def-in sh3addl cd2	r1	r2	t	6	02	16#56)
(def-in sh3addo cd2	r1	r2	t	6	02	16#76)
(def-in shd	cd5	r1	r2	p	t	10	16#34	16#2)
(def-in ssm	i	t	23	00	16#6b)
(def-in stb	r	d(s b)	1	16#18)
(def-in stbs	cp2	r	d (s b)	4	03	16#8)
(def-in stbys	r	d(s b)	4	03	16#c)
(def-in sth	r	d(s b)	1	16#19)
(def-in sths	cp2	r	d (s b)	4	03	16#9)
(def-in stw	r	d(s b)	1	16#1a)
(def-in stwas	cp2	r	d(b)	4	03	16#e)
(def-in stwm	r	d(s b)	1	16#1b)
(def-in stws	cp2	r	d (s b)	4	03	16#a)
(def-in sub	cd1	r1	r2	t	6	02	16#20)
(def-in subb	cd1	r1	r2	t	6	02	16#28)
(def-in subbo	cd1	r1	r2	t	6	02	16#68)
(def-in subi	cd1	im11	r	t	7	16#25 16#0)
(def-in subio	cd1	i	r	t	7	16#25 16#1)
(def-in subo	cd1	r1	r2	t	6	02	16#60)
(def-in subt	cd1	r1	r2	t	6	02	16#26)
(def-in subto	cd1	r1	r2	t	6	02	16#66)
(def-in sync	NIL	17	00	16#20)

(def-in uaddcm  cd4	r1	r2	t	6	02	16#4c)
(def-in uaddcmt cd4	r1	r2	t	6	02	16#4e)
(def-in uxor	cd4	r1	r2	t	6	02	16#1c)

(def-in vdep	cd5	r	len	t	9	16#35	16#1)
(def-in vdepi	cd5	im5	len	t	9	16#35	16#5)
(def-in vextrs  cd5	r	len	t	8	16#34	16#5)
(def-in vextru  cd5	r	len	t	8	16#34	16#4)
(def-in vshd	cd5	r1	r2	t	10	16#34	16#0)
 
(def-in xor	cd3	r1	r2	t	6	02	16#14)
 
(def-in zdep	cd5	r	p	len	t	9	16#35	16#2)
(def-in zdepi	cd5	im5	p	len	t	9	16#35	16#6)
(def-in zvdep	cd5	r	p	len	t	9	16#35	16#0)
(def-in zvdepi  cd5	im5	p	len	t	9	16#35	16#4)

%  Coprocessor

(def-in cldds	uid	cp2	i (s b)	t	31	16#0b)
(def-in clddx	uid	cp1	x (s b)	t	29	16#0b)
(def-in cldws	uid	cp2	d (s b)	t	31	16#09)
(def-in cldwx	uid	cp1	x (s b)	t	29	16#09)
(def-in copr	uid	sop	n 	nil	28	16#0c)
(def-in cstds	uid	cp2	r	d (s b)	32	16#0b)
(def-in cstdx	uid	cp1	r	x (s b)	30	16#0b)
(def-in cstws	uid	cp2	r	d (s b)	32	16#09)
(def-in cstwx	uid	cp1	r	x (s b)	30	16#09)

(def-in fabs	fmt	r	t	33	16#0c	16#3)
(def-in fadd	fmt	r1	r2	t	33	16#0c	16#0)
(def-in fcmp	ftm	cd6	r1	r2	35	16#0c	16#0)
(def-in fcnvff	sf	df	r	t	34	16#0c	16#0)
(def-in fcnvfx	sf	df	r	t	34	16#0c	16#2)
(def-in fcnvfxt	sf	df	r	t	34	16#0c	16#3)
(def-in fcnvxf	sf	df	r	t	34	16#0c	16#1)
(def-in fcpy	fmt	r	t	33	16#0c	16#2)
(def-in fdiv	fmt	r1	r2	t	36	16#0c	16#3)
(def-in fldds	cp2	d(s b)	t	31	16#0b	16#0)
(def-in flddx	cp1	x(s b)	t	29	16#0b	16#0)
(def-in fldws	cp2	d(s b)	t	31	16#09	16#0)
(def-in fldwx	cp1	x(s b)	t	29	16#09	16#0)
(def-in fmpy	fmt	r1	r2	t	36	16#0c	16#2)
(def-in fmpyadd f	rm1	rm2	tm	ra	ta	37	6)
(def-in fmpysub f	rm1     rm2     tm      ra      ta      37      16#26)
(def-in frem	fmt	r1	r2	t	36	16#0c	16#4)
(def-in frnd	fmt	r	t	33	16#0c	16#5)
(def-in fsqrt	fmt	r	t	33	16#0c	16#4)
(def-in fstds	cp2	r	d (s b)	32	16#0b	16#0)
(def-in fstdx	cp1	r	x (s b)	30	16#0b	16#0)
(def-in fstws	cp2	r	d (s b)	32	16#09	16#0)
(def-in fstwx	cp1	r	x (s b)	30	16#09	16#0)
(def-in fsub	fmt	r1	r2	t	36	16#0c	16#1)
(def-in ftest	nil	35	16#0c	16#1)

(def-in spop0	sfu	sop	n	nil	24	16#04)
(def-in spop1	sfu	sop	n	t	25	16#04)
(def-in spop2	sfu	sop	n	r	26	16#04)
(def-in spop3	sfu	sop	n	r1	r2	27	16#04)
 
(def-in xmpyu	fmt	r1	r2	t	35	16#0e	16#02)


%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
%k
%%      Assembler code generator
%%
%%|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

(compiletime 
(dm def-inst (u)
 (prog (name length result)
  (setq u (cdr u)) 
  (setq name (first u))
  (setq length (second u))
  (push (list 'put (mkquote name) (mkquote '*op-length*) length) result)
  (setq u (cddr u))    
  (setq result
     (append (mapcar u (function 
	 (lambda (x) (list 'put (mkquote name)
		                (mkquote (car x)) (mkquote (cdr x))))))
	result))
  (return(cons 'progn result))))
)

%%      Pseudo-Instructions

(def-inst  *ENTRY 0 (*apply* entry-gen))
%
%       (*ENTRY name type argn)
%
(de entry-gen (instruction name type argn)
  (saveentry (list instruction name type argn)))

%
%       (WORD value ......)
%
(def-inst WORD      4 (*apply* expand-word))
(def-inst FullWORD  4 (*apply* expand-word))
(def-inst STRING    0 (*apply* expand-string))
		            %will be treated by onelaplength
(def-inst Float     8 (*apply* expand-float))

(def-inst oddalign  0 (*apply* expand-oddalign))

(de expand-word (in x)        (store-word-expression x))

(de expand-string (in x)
	(for (from i 0 (size x))
	     (do (setf (byte (wplus2 codebase* currentoffset*) i) (indx x i))))
	(setf (byte (wplus2 codebase* currentoffset*) (wplus2 (size x) 1)) 0)
	(setq currentoffset* 
		(wplus2 (wtimes2 (strpack (size x))
				 addressingunitsperitem)
		 currentoffset*)))

(de expand-float (in x)         (store-double x ))

%%    The Instruction processor

(de expand-oddalign (in)
     (when (eq (wand currentoffset* (wnot 7)) currentoffset*)
	(setq currentoffset* (wplus2 currentoffset* addressingunitsperitem))))

(de depositinstruction (in)
   (prog (name format spec oldcurrent lng oldbto )
 
      (setq oldcurrent currentoffset*)
      (when *writingfaslfile (setq oldbto bittableoffset*))
      (when (atom in) (return nil)) % skip labels
      (setq name (first in))
      (setq lng (cond ((eq name 'string) (wtimes2 addressingunitsperitem
						(strpack (size (second in))) ))
		      ((eq name 'oddalign) 
			(if (eq currentoffset* (wand currentoffset* (wnot 7)))
				addressingunitsperitem 0))
		      (t (get name '*op-length*))))

      (when *writingfaslfile (updatebittable lng 0) % clear bittable 
		             (setq bittableoffset* oldbto))
      
      (cond ((get name '*apply*) (apply (car (get name '*apply*)) in))
		          % escape for special pseudo instructions
            ((setq format (get name 'format))
                         (format-matcher in (get name 'OPERANDS) nil format))
   (t (error 4711 (list "Wrong HP Risc instruction :" in))))
  % 
  % update currentoffset*
  %
    (setq currentoffset* (plus2 oldcurrent lng)) 
    (when *writingfaslfile (setq bittableoffset* (plus2 oldbto lng )))
) )

(de format-matcher (ins format initlist formfkt) 
   (prog (in name item fmtitem alist appl *lower)
       (setq alist 
	     (append initlist
       `((r1 . 0) (r2 . 0) (t . 0) (tgt . 0) (im5 . 0) (n . 0)
	(s . 0) (sr . 0) (cp . 0) (len . 0) (ext7 . 0) (cd1 . 0)
	(cd2 . 0) (cd3 . 0) (cd4 . 0) (cd5 . 0) (c . 0) 
	(m . 0) (u . 0) (e . 0) (ext4 . 0) (ext7 . 0) (ext3 . 0) (cp . 0)
	(clen . 0) (f . 0) (x . 0) (im14 . 0) (d16 . 0) )))
       (setq name (car ins))
       (setq alist (cons (cons 'op (get name 'opcode)) alist))
       (when (get name 'extension)
		(setq alist (cons (cons 'e (get name 'extension)) alist))
		(setq alist (cons (cons 'ext3 (get name 'extension)) alist))
		(setq alist (cons (cons 'ext4 (get name 'extension)) alist))
		(setq alist (cons (cons 'ext7 (get name 'extension)) alist))
		(setq alist (cons (cons 'ext8 (get name 'extension)) alist)))
       (setq formfkt (intern (bldmsg "%w-%d" 'formatfn formfkt)))
       (setq in (cdr ins)) 
	(when (and (get name 'suffixes) (pairp (car in)) 
		   (or (eq (caar in) 'compl) (eq (caar in) 'cond)))
		(setq alist (suffix-match (get name 'suffixes) (car in) alist))
		(setq in (cdr in)))
	(while (pairp in)
	 (setq item (car in))
	 (setq fmtitem (car format))
	 (cond  ((numberp fmtitem)  nil)
		((and (idp fmtitem) (setq appl (get fmtitem 'matchfn)))
		      (setq alist (apply appl (list (car in) alist))))
		(t (error 4711 (list "wrong HP Risc instruction " ins fmtitem
					))))
	 (setq in (cdr in))
	 (setq format (cdr format))
	 (when (and (pairp format) (pairp (car format)))
		(setq format (cdr format))) % skip (displacement ...
	 (when (and (not (pairp format)) in)
		(error 4711 (list "HP Risc wrong instruction :" ins))))

	(while format
	   (when (idp (car format))
	       (error 4711 (list "HP Risc instruction too short :" ins)))
	   (setq format (cdr format)))

   (apply formfkt (list name alist))
)  )
   
(de suffix-match (su in alist)
   (prog (suff item )
      (setq in (cdr in))
      (while  (and in su)
	   (setq suff (pop su))
           (setq item (pop in))
	   (cond ((eq suff 'cp1) (setq alist (cp1-match item alist)))
		 ((eq suff 'cp2) (setq alist (cp2-match item alist)))
		 ((eq suff 'cp3) (setq alist (cp3-match item alist)))
		 ((eq suff 'cp4) (setq alist (cp4-match item alist)))
		 ((eq suff 'fmt) (setq alist (fmt-match item alist)))
		 ((eq suff 'cd1) (setq alist (cd1-match item alist)))
		 ((eq suff 'cd2) (setq alist (cd2-match item alist)))
		 ((eq suff 'cd3) (setq alist (cd3-match item alist)))
		 ((eq suff 'cd5) (setq alist (cd5-match item alist)))
		 (t (error 4711 (list " no procedure for suffix" suff))))
      )
      (return alist)))


(de cp1-match (item alist)
    (prog (u m)
     (cond ((eq item 'nil) (setq u 0 m 0))
	   ((eq item 'M)   (setq u 0 m 1))
	   ((eq item 'S)   (setq u 1 m 0))
	   ((eq item 'SM)  (setq u 1 m 1)))
     (return (append `((u . ,u) (m . ,m)) alist))))
       
(de cd1-match (item alist)
    (prog (c f)
     (cond ((eq item 'nil) (setq c 0 f 0))
           ((eq item '=)   (setq c 1 f 0))
	   ((eq item '<)   (setq c 2 f 0))
	   ((eq item '<=)  (setq c 3 f 0))
	   ((eq item '<<)  (setq c 4 f 0))
	   ((eq item '<<=) (setq c 5 f 0))
	   ((eq item 'SV)  (setq c 6 f 0))
	   ((eq item 'OD)  (setq c 7 f 0))
	   ((eq item 'TR ) (setq c 0 f 1))
           ((eq item '<>)  (setq c 1 f 1))
           ((eq item '>=)  (setq c 2 f 1))
           ((eq item '>)   (setq c 3 f 1))
           ((eq item '>>=) (setq c 4 f 1))
           ((eq item '>> ) (setq c 5 f 1))
           ((eq item 'NSV) (setq c 6 f 1))
           ((eq item 'EV)  (setq c 7 f 1)))
     (return (cons (cons 'c c) (cons (cons 'f f) alist)))))

(de cd2-match (item alist)
    (prog (c f)
     (cond ((eq item 'nil) (setq c 0 f 0))
           ((eq item '=)   (setq c 1 f 0))
           ((eq item '<)   (setq c 2 f 0))
           ((eq item '<=)  (setq c 3 f 0))
           ((eq item 'NUV) (setq c 4 f 0))
           ((eq item 'ZNV) (setq c 5 f 0))
           ((eq item 'SV)  (setq c 6 f 0))
           ((eq item 'OD)  (setq c 7 f 0))
           ((eq item 'TR)  (setq c 0 f 1))
           ((eq item '<>)  (setq c 1 f 1))
           ((eq item '>=)  (setq c 2 f 1))
           ((eq item '>)   (setq c 3 f 1))
           ((eq item 'UV)  (setq c 4 f 1))
           ((eq item 'VNZ) (setq c 5 f 1))
           ((eq item 'NSV) (setq c 6 f 1))
           ((eq item 'EV)  (setq c 7 f 1)))
     (return (cons (cons 'c c) (cons (cons 'f f) alist)))))

(de cd3-match (item alist)
    (prog (c f)
     (cond ((eq item 'nil) (setq c 0 f 0))
           ((eq item '=)   (setq c 1 f 0))
           ((eq item '<)   (setq c 2 f 0))
           ((eq item '<=)  (setq c 3 f 0))
           ((eq item 'OD)  (setq c 7 f 0))
           ((eq item 'TR)  (setq c 0 f 1))
           ((eq item '<>)  (setq c 1 f 1))
           ((eq item '>=)  (setq c 2 f 1))
           ((eq item '>)   (setq c 3 f 1))
           ((eq item 'EV)  (setq c 7 f 1)))
     (return (cons (cons 'c c) (cons (cons 'f f) alist)))))

(de cd5-match (item alist)
    (prog (c )
     (cond ((eq item 'nil) (setq c 0))
           ((eq item '=)   (setq c 1))
           ((eq item '<)   (setq c 2))
           ((eq item 'OD)  (setq c 3))
           ((eq item 'TR)  (setq c 4))
           ((eq item '<>)  (setq c 5))
           ((eq item '>=)  (setq c 6))
           ((eq item 'EV)  (setq c 7)))
     (return (cons (cons 'c c) alist))))

(put 'im5 'matchfn 'im5-match)

(de im5-match (x alist)
    (if (and (numberp x) (lessp x 16) (greaterp x -17)) 
		(cons (cons 'r!/im5 (low_sign_5 x))
		   (cons (cons 'r1!/im5 (low_sign_5 x))
			(cons (cons 'im5  (low_sign_5 x)) alist)))
     (matcherror "im5" x)))

(put 'im11 'matchfn 'im11-match)

(de im11-match (x alist)
    (if (and (numberp x) (lessp x 16#400) (greaterp x (minus 16#401)))
                (cons (cons 'im11 (low_sign_11 x)) alist)
     (matcherror "im11" x)))

(put 'tgt 'matchfn 'tgt-match)

(de tgt-match (tt alist)
    (setq tt (jumpaddressvalue tt))
    (if (or (eq (ashift (ashift (wdifference tt 8) -2) -11) -1)
	    (eq (ashift (wshift (wdifference tt 8) -2) -11) 0))
	(append (assemble_12 (wshift (wdifference tt 8) -2)) alist)
	(error 4711 " Compiler bug (long cond jumps)")))

(put 'target 'matchfn 'target-match)

(de target-match (t alist)
    (setq t (jumpaddressvalue t))
    (append (assemble_17 (wshift (wdifference t 8) -2)) alist))

(put 'im21 'matchfn 'im21-match)

(de im21-match (x alist)
    (setq x (addressvalue x))
    %(if (and (numberp x) (lessp x 16#10000) (greaterp x (minus 16#10001)))
                (cons (cons 'im21 (assemble_21 x)) alist)
)
     %(matcherror "im21" x)))

(put 't 'matchfn 't-match)

(de t-match (r alist)
    (if (regp r) (cons (cons 't!/r r) (cons (cons 't r) alist))
     (matcherror "Register" r)))

(put 'r 'matchfn 'r-match)

(de r-match (x alist)
    (if (regp x) (cons (cons 'r!/im5 x)
		 (cons (cons 'r1!/im5 x)
		 (cons (cons 't!/r x) (cons (cons 'r x) alist))))
     (matcherror "Register" x)))

(put 'r1 'matchfn 'r1-match)

(de r1-match (x alist)
    (if (regp x) (cons (cons 'r1!/im5 x) 
		(cons (cons 'r1 x) (cons (cons 'r!/im5 x) alist)))
     (matcherror "Register" x)))

(put 'r2 'matchfn 'r2-match)

(de r2-match (x alist)
    (if (regp x)  (cons (cons 'r2!/p x) (cons (cons 'r2 x) alist))
     (matcherror "Register" x)))

(put 'p 'matchfn 'p-match)

(de p-match (x alist)
    (if (and (posintp x) (lessp x 32)) 
		(cons (cons 'p x) 
		  (cons (cons 'r2!/p x)
		    (cons (cons 'cp (wdifference 31 x)) alist)))
     (matcherror "Position" x)))

(put 'len 'matchfn 'len-match)

(de len-match (x alist)
    (if (and (posintp x) (lessp x 33))
         	(cons (cons 'clen  (wdifference 32 x)) alist)
     (matcherror " Length " x)))

(put 'x 'matchfn 'indexed-match)

(de indexed-match (x alist)
      (cond ((eq (car x) 'deferred)
                 (cons (cons 'immediate 0) (cons (cons 'r1 (cadr x)) alist)))
            ((and (eq (car x) 'displacement2) (regp (cadr x)) (regp (caddr x)))
                     (cons (cons 'x (caddr x))
			(cons (cons 't!/b (cadr x))
                           (cons (cons 'b (cadr x)) alist))))
            ((and (eq (car x) 'indexed) (regp (cadr x))
			 (regp (caddr x)) (regp (cadddr x)))
                      (cons (cons 'x (cadr x))
                        (cons (cons 's 
				(assemble_3 (get-register-code (caddr x))))
			 (cons (cons 'sr
				(assemble_3 (get-register-code (caddr x))))
                           (cons (cons 'b (cadddr x)) alist)))))
            (t (matcherror "indexed" x))))

(put 'd 'matchfn 'displacement-match)

(de displacement-match (x alist)
      (cond ((eq (car x) 'deferred)
		 (cons (cons 'immediate 0) (cons (cons 'r1 (cadr x)) alist)))
	    ((eq (car x) 'displacement2)
		     (cons (cons 'im14 (low_sign_14 (addressvalue (caddr x))))
		           (cons (cons 'b (cadr x)) alist)))
            ((eq (car x) 'displacement)
                     (cons (cons 'im14 (low_sign_14 (addressvalue (cadddr x))))
			(cons (cons 's (assemble_3 
					(get-register-code (cadr x))))
			 (cons (cons 'sr (assemble_3
					(get-register-code (cadr x))))
                           (cons (cons 'b (caddr x)) alist)))))
	    ((eq (car x) 'entryaddress)
		(cons (cons 'im14
		        (wand (addressvalue (list 'xidloc (cadr x))) 16#ffff))
		        (cons (cons 'r1 '(reg symfncstart)) alist)))
	    ((eq (car x) 'valueaddress)
		(cons (cons 'immediate 
		        (wand (addressvalue (list 'xidloc (cadr(cadr x))))
		                 16#ffff))
		        (cons (cons 'r1 '(reg symvalstart)) alist)))
	    (t (matcherror "based" x))))

(put 'wd 'matchfn 'worddisplacement-match)

(de worddisplacement-match (x alist)
   (let (w)
      (cond ((eq (car x) 'deferred)
                 (cons (cons 'immediate 0) (cons (cons 'r1 (cadr x)) alist)))
            ((eq (car x) 'displacement)
		    (setq w (addressvalue (cadddr x)))
		    (setq alist (append (assemble_17 (wshift w -2)) alist))
                    (cons (cons 's (assemble_3
                                        (get-register-code (cadr x))))
			(cons (cons 'sr (assemble_3
					(get-register-code (cadr x))))
			  (cons (cons 'b (caddr x)) alist))))
            (t (matcherror "(wd) indexed" x)))))

(de matcherror (expected got)
    (error 4711 (list "Mismatch of Arguments (" expected "expected) :" got))) 

(def-inst nop 4  (*apply* nop-gen))

(de nop-gen () (depositinstruction '(or (reg r0) (reg r0) (reg r0))))

(def-inst comb 4  (*apply* comb-gen))

(def-inst comib 4  (*apply* comib-gen))

(de comb-gen (name suff imm r tgt)
      (cond ((memq (cadr suff) '(NIL = < <= << <<= SV OD))
		   (depositinstruction `(combt ,suff ,imm ,r ,tgt)))
            ((memq (cadr suff) '(TR <> >= > >>= >> NSV EV))
		   (depositinstruction `(combf ,suff ,imm ,r ,tgt)))
	    (t (error 99 (list "wrong comb" suff imm r tgt)))))

(de comib-gen (name suff imm r tgt)
      (cond ((memq (cadr suff) '(NIL = < <= << <<= SV OD))
                   (depositinstruction `(comibt ,suff ,imm ,r ,tgt)))
            ((memq (cadr suff) '(TR <> >= > >>= >> NSV EV))
                   (depositinstruction `(comibf ,suff ,imm ,r ,tgt)))
            (t (error 99 (list "wrong comib" suff imm r tgt)))))

(def-inst mtsar 4 (*apply* mtsar-gen))

(de mtsar-gen (name r)
     (depositinstruction `(mtctl ,r (reg cr11))))

%%      move-relocatable-object pseudo instructions

(def-inst move-relocatable-object 8 (*apply* movex-gen))

(de movex-gen (name src reg)
	  (cond (*writingfaslfile
	     (store-word-expression src) %% when writing to faslfile, generate
	     (relocatable-movex-gen reg)) %% special code for relocation.
	   (t (depositinstruction `(ldil (high ,src) ,reg)) 
		                         %% otherwise normal code.
	      (depositinstruction `(ldo  (displacement2 ,reg (low ,src)) ,reg)))))

(def-inst  move-relocatable-object-and-load 8 (*apply* movex-load))

(de movex-load (name src reg)
   (cond (*writingfaslfile
	   (store-word-expression  src)  %% when writing to faslfile, generate
	   (relocatable-movex-load  reg))   %% special code for relocation.
	 (t (depositinstruction `(ldil (high ,src) ,reg))
		                 %% otherwise normal code.
	    (depositinstruction `(ldw (displacement (reg sr0),reg (low ,src)) ,reg)))))

(def-inst move-relocatable-object-and-store 8 (*apply* movex-store))

(de movex-store (name reg cont dst)
  (cond (*writingfaslfile
	   (store-word-expression dst) %% when writing to faslfile, generate
	   (relocatable-movex-store reg cont)) %% code for relocation.
	(t (depositinstruction `(ldil (high ,dst) ,cont))
		                          %% otherwise normal code.
	   (depositinstruction 
		`(stw ,reg (displacement (reg sr0) ,cont (low ,dst)))))))

(de relocatable-movex-gen (reg)
  (let ((regcode (get-register-code reg)) (mov-instr 16#0d))
       (store-relocatable-movex            % ldo instr 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21) (wshift regcode 16))))))

(de relocatable-movex-store (reg cont)
  (let ((regcode (get-register-code reg))
	(regcode1 (get-register-code cont))
	(mov-instr 16#1a))  %stw
       (store-relocatable-movex 
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode1 21) (wshift regcode 16))))))

(de relocatable-movex-load (reg)
  (let ((regcode (get-register-code reg)) (mov-instr 16#12))
       (store-relocatable-movex                      % ldw
	    (wplus2 (wshift mov-instr 26)
		    (wplus2  (wshift regcode 21) (wshift regcode 16)))))) 

(de assemble_21 (x)
    (prog (b20 b9_19 b5_6 b0_4 b7_8)
          (setq b20 (wshift x -20))
          (setq b9_19 (wand (wshift x -9) 16#7ff))
          (setq b5_6  (wand (wshift x -7) 3))
          (setq b0_4  (wand (wshift x -2) 16#1f))
          (setq b7_8  (wand x 3))
          (return
           (wor (wshift b0_4 16)
                (wor (wshift b5_6 14)
                    (wor (wshift b7_8 12)
                        (wor (wshift b9_19 1) b20)))))))

(de assemble_17 (x)
    (prog (w w1 w2)
        (setq w (wand (wshift x -16) 1))
        (setq w1 (wand (wshift x -11) 16#1f))
        (setq w2 (wor (wand (wshift x -10) 1) (wand (wshift x 1) 16#7fe)))
        (return (cons (cons 'w w) 
			(cons (cons 'w1 w1) (cons (cons 'w2 w2) nil))))))

(de assemble_3(x)
    (wor (wand (wshift x -2) 1) (wand (wshift x 1) 2#110)))

(de assemble_12 (x)
    (prog (w w1)
	(setq w (wand (wshift x -11) 1))
	(setq w1 (wor (wand (wshift x -10) 1) (wand (wshift x 1) 16#7fe)))
        (return (cons (cons 'w w) (cons (cons 'w1 w1) nil)))))

(de low_sign_14 (x)
     (wor (wand (wshift x -13) 1) (wand (wshift x 1) 16#3fff)))

(de low_sign_5 (x)
     (wor (wand (wshift x -4) 1) (wand (wshift x 1) 16#1f)))

(de low_sign_11 (x)
     (wor (wand (wshift x -10) 1) (wand (wshift x 1) 16#7ff)))

