%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXC:HP-COMP.SL
% Description:  Compiler patterns for HP9836 PSL, plus a few CMACRO expanders
% Author:       G.Q. Maguire, M.L. Griss and Sam Sands
% Created:      16-Jan-1982
% Modified:     29-Oct-84 08:55:07 (Vicki O'Day)
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 25-Jul-86 (Leigh Stoller)
%  Added a second case to the memory pattern that makes sure the simpler arg
%  is resolved second to the temp register is not wasted.
% 24-Jul-86 (Leigh Stoller)
%  Put in a new version of the MEMORY pattern that returns itself. The pattern
%  was highly dependent on the 14 real reg mode.
% 21-Jul-86 (Bob Kessler & Leigh Stoller)
%  Put in a new version of the carcdrpat pattern that starts the argument in
%  a Dreg instead of an Areg. 
% 08-Feb-84 1:00:00  (Sam Sands)
%  Corrected function &ALTDEST. Two-operand operators (e.g., eq, neq, wleq,
%   etc.
%  when their operands used a register, the new compiler "stepped on" the
%  register.
% 15-Dec-83 12:49:19 (Tim Tillson)
%  Added header to Sam Sands' D-register version
% 29 Mar 1983 1135-PST (Cris Perdue)
%  Added a USESDEST case to the pattern for SUBPAT
% 28-Mar-83 14:01:59, KENDZIERSKI
%  Fixed header filename/comments to be correct and in proper chron. order.
% 22-Oct-82 14:02:56, BENSON
%  Changed ASSOCPAT to be the same as the 20 and Vax
% 31-Jan-82 11:54:12, GRISS
%  Improved ForeignLink
% 28-Jan-82 13:20:47, GRISS
%  Added INFPATs and TAGPATs and changed OPENFN's
% 18-Jan-82 06:47:49, GRISS
%  Fixed the setup of MoveSeq and PushSeq, changed to CONS from Append
% 18-Jan-82 05:17:37, GRISS
%  replace (MEMORY argn) by (WVAR argn) in Foreign
% 18-Jan-82 04:56:03, GRISS
%  Fixed typo Seqence to Sequence
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%(PUT 'CARCDRPAT 'PATTERN                %  CAR, CDR              scs 
% '(&ALTAREG nil 
%    
%   ((AREG)                   ('*RETURN (FN A1))) 
%   ( ANY     ('*LOAD TA1 A1) ('*RETURN (FN TA1)))))


% The above version was used in the older 68000 model that did not need tag
%  stripping when doing car and cdr's. It was okay to put the argument directly
%  into an A reg for memory indirection. Now that we do tag stripping, we want
%  the arg to start in a Dreg to avoid having it moved by *field back into
%  a Dreg from an Areg. A waste of instrucions. /LBS

(PUT 'CARCDRPAT 'PATTERN                %  CAR, CDR
 '(&ALTAREG nil 
    
   ((DREG)		     ('*RETURN (FN A1))) 
   ( ANY     ('*LOAD TD1 A1) ('*RETURN (FN TD1)))))

   
(commentoutcode
(PUT 'MEMORYPAT 'PATTERN                %  MEMORY                scs 
 '(&ALTAREG nil 
    
   ((AREG   DISP)  ('*RETURN (FN A1 A2))) 
   ((DISP   AREG)  ('*RETURN (FN A2 A1))) 
   ((AREG   REG)   ('*RETURN (FN A1 A2))) 
   ((REG    AREG)  ('*RETURN (FN A2 A1))) 
   ((AREG   ANY)   ('*WPLUS2 A1 A2) ('*SET A1 ('*WPLUS2 A1 A2)) 
                                    ('*RETURN (FN A1 ('WCONST '0)))) 
   ((ANY    AREG)  ('*WPLUS2 A2 A1) ('*SET A2 ('*WPLUS2 A2 A1)) 
                                    ('*RETURN (FN A2 ('WCONST '0))))
   ((ANY    DISP)  ('*LOAD TA1 A1)  ('*RETURN (FN TA1 A2))) 
   ((DISP   ANY)   ('*LOAD TA2 A2)  ('*RETURN (FN TA2 A1))) 
   
   ((ANY    DREG)  ('*LOAD TA1 A1)  ('*RETURN (FN TA1 A2))) 
   ((DREG   ANY)   ('*LOAD TA2 A2)  ('*RETURN (FN TA2 A1))) 
   ( ANY           ('*LOAD TA1 A1)  ('*WPLUS2 TA1 A2) 
		   ('*SET TA1 ('*WPLUS2 A1 A2)) 
                   ('*RETURN (FN TA1 ('WCONST '0))))))
)

% This one lets the cmacro make most of the decisions about the expansion
%  of its arguments. One case that is needed is if the first argument is
%  a reg. We do not want the one temp register allocated for memory to be
%  wasted on resolving a register, when the second arg might need it.

(PUT 'MEMORYPAT 'PATTERN      % MEMORY
 '(&ALTAREG nil 
   ((Disp  Reg)     ('*RETURN (FN A2 A1)))
   ((Reg   Openfn)  ('*RETURN (FN A2 A1)))
   ((Any   Any)     ('*RETURN (FN A1 A2)))))


(PUT 'LOCPAT 'PATTERN '(NIL
  ('*SET DEST (FN A1))
  ((NOTIMM  DDEST)   (MAC TA2  A1)  ('*RETURN TA2))
  ( ANY              (MAC DEST A1))
  ))


(PUT 'ADDPAT 'PATTERN '(nil             %  WPLUS2               scs  
   ('*SET DEST (FN A1 A2)) 
   ((DEST     MOVEQ) ('*LOAD TD2 A2)  (MAC DEST TD2)) 
   ((MOVEQ    DEST)  ('*LOAD TD1 A1)  (MAC DEST TD1)) 
   ((USESDEST MOVEQ) ('*LOAD TD2 A2)  (MAC TD2 A1) ('*RETURN TD2)) 
   ((MOVEQ USESDEST) ('*LOAD TD1 A1)  (MAC TD1 A2) ('*RETURN TD1)) 
   ((ANY      MOVEQ) ('*LOAD DEST A2) (MAC DEST A1)) 
   ((MOVEQ    ANY)   ('*LOAD DEST A1) (MAC DEST A2)) 
   ((ANY      DEST)  (MAC A2 A1)) 
   ((REGNP    ANY)   (MAC A1 A2) ('*RETURN A1)) 
   ((ANY      REGNP) (MAC A2 A1) ('*RETURN A2)) 
   ((USESDEST 
           USESDEST) ('*LOAD T1   A1) (MAC T1 A2) ('*RETURN T1)) 
   ((ANY   USESDEST) ('*LOAD DEST A2) (MAC DEST A1)) 
   ((CONST ANY)      ('*LOAD DEST A2) (MAC DEST A1)) 
   ( ANY             ('*LOAD DEST A1) (MAC DEST A2))))
   

(PUT 'SUBPAT 'PATTERN '(nil             %  WDIFFERENCE           scs
   ('*SET DEST (FN A1 A2)) 
   ((REGNP MOVEQ)    ('*LOAD TD1 A2)  (MAC A1  TD1)   ('*RETURN A1)) 
   ((REGNP ANY)                       (MAC A1  A2)    ('*RETURN A1)) 
   ((DISPL USESDEST) ('*LOAD TA1 A1)  (MAC TA1 A2)    ('*RETURN TA1)) 
   ((ANY USESDEST)   ('*LOAD T1 A1)   (MAC T1 A2)     ('*RETURN T1)) 
   ((ANY MOVEQ)      ('*LOAD DEST A1) ('*USES DEST)   
                     ('*LOAD TD2  A2) (MAC DEST TD2)) 
   ( ANY             ('*LOAD DEST A1)                 (MAC DEST A2))))
   

(PUT 'UNARYPAT 'PATTERN '(nil           %  WMINUS, WNOT           scs
   ('*SET DEST (FN A1)) 
   ((DREGNP)         (MAC A1  A1) ('*RETURN A1)) 
   (ANY              (MAC TD1 A1) ('*RETURN TD1))))
   

                                    %  WAND, WOR, WXOR                scs
                                    %  TEMPORARY INEFFICIENT VERSION  scs
(PUT 'ANDORXORPAT 'PATTERN '(nil 
   ('*SET DEST (FN A1 A2)) 
   ((DREG   DREGDEST)                 (MAC A2 A1)) 
   ((DREGNP DREG)                     (MAC A1 A2)  ('*RETURN A1)) 
   ((DREG   DREGNP)                   (MAC A2 A1)  ('*RETURN A2)) 
   ((DREGNP ANY)      ('*LOAD TD2 A2) (MAC A1 TD2) ('*RETURN A1)) 
   ((ANY    DREGNP)   ('*LOAD TD1 A1) (MAC A2 TD1) ('*RETURN A2)) 
   ((ANY    ANY)      ('*LOAD TD1 A1) 
                      ('*LOAD TD2 A2) (MAC TD1 TD2) ('*RETURN TD1))))
   
(commentoutcode
(PUT 'SETQPAT 'PATTERN '(nil nil        %  SETQ
   
   (      (ANY  REG)    ('*STORE A2 A1)                   ('*RETURN A2)) 
   (      (ANY  SMALL)  ('*LOAD TD2 A2)  ('*STORE TD2 A1) ('*RETURN TD2)) 
   ((NOVAL ANY  ANY)    ('*STORE A2 A1)) 
   (  (USESDEST ANY)    ('*LOAD TD2 A2)  ('*STORE TD2 A1) ('*RETURN TD2)) 
   ( ANY                ('*LOAD DEST A2) ('*STORE DEST A1))))
   

(PUT 'RPLACPAT 'PATTERN                 %  REPLACA, REPLACD
   '(&ALTAREG 
    NIL
   ((AREG ZERO)         ('*STORE A2  (MAC A1))            ('*RETURN A1 )) 
   ((AREG SMALL)                           ('*LOAD TD2 A2) 
                        ('*STORE TD2 (MAC A1))            ('*RETURN A1 )) 
   ((AREG ANY)          ('*STORE A2  (MAC A1))            ('*RETURN A1 )) 

   ((ANY  ZERO  DDEST)  ('*LOAD  TA1 A1)    
                        ('*STORE A2 (MAC TA1))            ('*RETURN TA1)) 
   ((ANY  SMALL DDEST)  ('*LOAD  TA1 A1)   ('*LOAD TD2 A2) 
                        ('*STORE TD2 (MAC TA1))           ('*RETURN TA1)) 

   ((ANY  ANY   DDEST)  ('*LOAD  TA1 A1)    
                        ('*STORE A2 (MAC TA1))            ('*RETURN TA1)) 

   ((ANY  ZERO)         ('*LOAD  DEST A1)    
                        ('*STORE A2  (MAC DEST))) 
   ((ANY  SMALL)        ('*LOAD  DEST A1)   ('*LOAD TD2 A2) 
                        ('*STORE TD2 (MAC DEST))) 

   ((ANY  ANY)          ('*LOAD  DEST A1)    
                        ('*STORE A2  (MAC DEST))) 

   ))
)

(PUT 'SETQPAT 'PATTERN '(NIL NIL

 ((NOVAL ANY NOTANYREG) ('!*STORE A2 A1))
 ((NOVAL DEST ANY) 	('!*STORE A2 DEST))
 ((NOVAL USESDEST ANY) 	('!*LOAD T1 A2) 
			('!*STORE T1 A1))
 ((NOVAL ANY ANY) 	('!*LOAD DEST A2) 
			('!*STORE DEST A1))
 ((ANY DEST) 		('!*STORE DEST A1))
 ((DEST ANY) 		('!*STORE A2 DEST))
 ((USESDEST ANY) 	('!*STORE A2 A1) 
			('!*STORE A2 DEST))
 (ANY 			('!*LOAD DEST A2) 
			('!*STORE DEST A1))))

(PUT 'RPLACPAT 'PATTERN '(NIL NIL

   ((NOVAL ANY ANY) 	('!*STORE A2 (MAC A1)))
   ((DEST ANY) 		('!*STORE A2 (MAC A1)))
   ((USESDEST ANY) 	('!*STORE A2 (MAC A1)) 
			('!*LOAD DEST A1))
   ((ANY DEST) 		('!*STORE A2 (MAC A1)) 
			('!*LOAD DEST A1))
   ((ANY USESDEST) 	('!*STORE A2 (MAC A1)) 
			('!*LOAD DEST A1))
   (ANY 		('!*LOAD DEST A1) 
			('!*STORE A2 (MAC DEST)))))

   


(PUT 'SHIFTPAT 'PATTERN '(nil           %  WSHIFT, ASHIFT    scs
   NIL
   ((DREGNP CONST)       (MAC A1  A2) ('*RETURN A1)    ('*SET A1  (FN A1 A2))) 
   ((AREGNP LEQ2)        (MAC A1  A2) ('*RETURN A1)    ('*SET A1  (FN A1 A2))) 
   ((ANY    CONST)       ('*LOAD  TD1 A1) (MAC TD1 A2) ('*RETURN TD1) 
                                                       ('*SET TD1 (FN A1 A2))) 
   ((DREGNP DREGNP)      (MAC A1  A2)  ('*RETURN A1)   ('*SET A1  (FN A1 A2))
                         ('*DESTROY A2)) 
   ((ANY    DREGNP)      ('*LOAD  TD1 A1) 
                         (MAC TD1 A2)  ('*RETURN TD1)  ('*SET TD1 (FN A1 A2))
			 ('*DESTROY A2)) 
   ((DREGNP ANY)         ('*LOAD  TD2 A2) 
                         (MAC A1  TD2) ('*RETURN A1)   ('*SET A1  (FN A1 A2))
			 ('*DESTROY TD2)) 
   ((ANY    ANY)         ('*LOAD  TD1 A1) ('*LOAD TD2 A2) 
                         (MAC TD1 TD2) ('*RETURN TD1)  ('*SET TD1 (FN A1 A2))
			 ('*DESTROY TD2))
   ))
   

(PUT 'MKITEMPAT 'PATTERN '(nil          %   MKITEM   scs
   ('*SET DEST (FN A1 A2)) 
   ((DREGNP CONST)                         (MAC A1   A2) ('*RETURN A1)) 
   ((ANY    CONST DDEST)  ('*LOAD DEST A1) (MAC DEST A2)) 
   ((ANY    CONST)        ('*LOAD TD1  A1) (MAC TD1  A2) ('*RETURN TD1)) 
   ((REGNP  DREG)                          (MAC A1   A2) ('*RETURN A1)) 
   ((ANY    DREGDEST)     ('*LOAD TD1  A1) (MAC TD1  A2) ('*RETURN TD1)) 
   ((ANY    DREG  DDEST)  ('*LOAD DEST A1) (MAC DEST A2)) 
   ((REGNP  ANY)          ('*LOAD TD2  A2) (MAC A1   TD2) ('*RETURN A1)) 
   ((ANY    ANY   DDEST)  ('*LOAD DEST A1) ('*USES   DEST)
			  ('*LOAD TD2  A2) (MAC DEST TD2)) 
   ( ANY                  ('*LOAD TD1  A1) 
			  ('*LOAD TD2  A2) (MAC TD1  TD2) ('*RETURN TD1))))
   


                 
(PUT 'PUTFIELDPAT 'PATTERN '(nil        %  PUTFIELD          scs
 NIL
 ((CONST  DREGNP ANY ANY) (MAC A1  A2  A3 A4)                 ('*RETURN A2)
                          ('*SET DEST (FN A1 A2 A3 A4)) ) 
 ((DREGNP DREGNP ANY ANY) (MAC A1  A2  A3 A4) ('*RETURN A2)  
                          ('*SET DEST (FN A1 A2 A3 A4)) ('*DESTROY A1) ) 
 ((ANY    DREGNP ANY ANY) ('*LOAD  TD1 A1) 
                          (MAC TD1 A2  A3 A4) ('*RETURN A2) 
			  ('*SET DEST (FN A1 A2 A3 A4)) ('*DESTROY TD1) ) 
 ((CONST  ANY  ANY DDEST) ('*LOAD DEST A2)    (MAC A1 DEST A3 A4)
                          ('*SET DEST (FN A1 A2 A3 A4)) ) 
 ((CONST  ANY    ANY ANY) ('*LOAD TD2  A2)    (MAC A1 TD2 A3 A4) ('*RETURN TD2)
                          ('*SET DEST (FN A1 A2 A3 A4)) ) 
 ((DREGDEST ANY  ANY ANY) ('*LOAD TD2  A2) 
                          (MAC A1 TD2  A3 A4) ('*RETURN TD2) 
			  ('*SET DEST (FN A1 A2 A3 A4)) ('*DESTROY A1) ) 
 ((DREGNP  ANY ANY DDEST) ('*LOAD DEST A2) 
                          (MAC A1 DEST A3 A4) 
			  ('*SET DEST (FN A1 A2 A3 A4)) ('*DESTROY A1) ) 
 ((DREGNP  ANY ANY ANY)   ('*LOAD TD2  A2) 
                          (MAC A1 TD2  A3 A4) ('*RETURN TD2)
			  ('*SET DEST (FN A1 A2 A3 A4)) ('*DESTROY A1) ) 
 ((ANY     ANY ANY DDEST) ('*USES DEST)       ('*LOAD TD1  A1)('*LOAD DEST A2) 
                          (MAC TD1 DEST A3 A4) 
			  ('*SET DEST (FN A1 A2 A3 A4)) ('*DESTROY TD1) ) 
 ( ANY                    ('*LOAD TD1  A1)    ('*LOAD TD2 A2) 
                          (MAC TD1 TD2 A3 A4) ('*RETURN TD2) 
			  ('*SET DEST (FN A1 A2 A3 A4)) ('*DESTROY TD1) )
 ))
 


(PUT 'FIELDPAT 'PATTERN '(nil           %  FIELD, SIGNEDFIELD   scs
   ('*SET DEST (FN A1 A2 A3)) 
   ((DREGNP ANY ANY)       (MAC A1   A1 A2 A3) ('*RETURN A1)) 
   ((ANY    ANY ANY DDEST) (MAC DEST A1 A2 A3)) 
   ( ANY                   (MAC TD1  A1 A2 A3) ('*RETURN TD1))))
   

(PUT 'ADDMEMPAT 'PATTERN '(nil nil      %  ADDMEM          scs
   
   ((NOVAL ANY MOVEQ)  ('*LOAD    TD2  A2) (MAC     A1   TD2)) 
   ((NOVAL ANY CONST)                      (MAC     A1   A2)) 
   ((NOVAL ANY DREG)                       (MAC     A1   A2)) 
   ((NOVAL ANY ANY)    ('*LOAD    TD2  A2) (MAC     A1   TD2)) 
   (      (ANY REGNP)  ('*WPLUS2  A2   LA1) ('*STORE A2   A1) ('*RETURN A2)) 
   ( (USESDEST ANY)    ('*LOAD    T2   A2) 
		       ('*WPLUS2  T2   LA1) ('*STORE T2   A1) ('*RETURN T2)) 
   ( ANY               ('*LOAD    DEST A2) 
		       ('*WPLUS2  DEST LA1) ('*STORE DEST A1))
   ))
   

(PUT 'SUBMEMPAT 'PATTERN '(nil nil      %  SUBMEM          scs
   
   ((NOVAL ANY MOVEQ)  ('*LOAD  TD2 A2)    (MAC     A1   TD2)) 
   ((NOVAL ANY CONST)                      (MAC     A1   A2)) 
   ((NOVAL ANY DREG)                       (MAC     A1   A2)) 
   ((NOVAL ANY ANY)    ('*LOAD    TD2  A2) (MAC     A1   TD2)) 
   ( (ANY  USESDEST)   ('*LOAD        T1   LA1) 
		       ('*WDIFFERENCE T1   A2)  ('*STORE T1   A1) ('*RETURN T1)) 
   (       ANY MOVEQ   ('*LOAD        DEST LA1) ('*USES DEST)
	               ('*LOAD        TD2  A2)
		       ('*WDIFFERENCE DEST TD2) ('*STORE DEST A1))
   (       ANY         ('*LOAD        DEST LA1) 
		       ('*WDIFFERENCE DEST A2)  ('*STORE DEST A1))
   ))
   

% PUT('MODMEMPAT1,'PATTERN,'(                      %  ???
%   NIL NIL
%    (ANY (MAC A1))));


(PUT 'GETUNSBYTE 'PATTERN '(nil 
  
  ('*SET DEST (FN A1)) 
  ((USESDEST) (MAC A1 TD2) ('*RETURN TD2)) 
  ((DDEST)    (MAC A1 DEST)) 
  ( ANY       (MAC A1 TD2) ('*RETURN TD2))))
  

(PUT 'PUTANYBYTE 'PATTERN '(nil nil 
  
  ((ANY DREG)                     (MAC A2  A1)) 
  ( ANY           ('*LOAD TD2 A2) (MAC TD2 A1))))
  

(PUT 'GETSGNHWORD 'PATTERN '(nil 
  
  ('*SET DEST (FN A1)) 
  ( ANY       (MAC A1 DEST))))
  

(PUT 'PUTANYHWORD 'PATTERN '(nil nil 
  
  
  ((ANY REG)                     (MAC A2 A1)) 
  ( ANY           ('*LOAD T2 A2) (MAC T2 A1))))
  

(PUT 'TVPAT 'PATTERN                    %  EQ, NE, WLEQ, etc FN
  '(&ALTDEST 
  ('*SET DEST (FN A1 A2)) 
% ((DEST   DEST)   SHOULD BE CAUGHT BY PASS 1
  ((AFRAME ZERO)   ('*USES DEST)       ('*LOAD TD1  A1) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 TD1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((ZERO   AFRAME) ('*USES DEST)       ('*LOAD TD2  A2) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 A1 TD2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    AREG)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((AREG   ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    ZERO)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((ZERO   ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((SMALL  ANY)    ('*USES DEST)       ('*LOAD TD1  A1) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 TD1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((ANY    SMALL)  ('*USES DEST)       ('*LOAD TD2  A2) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 A1 TD2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    DREG)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((DREG   ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ((ANY    CONST)  ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 
  ((CONST  ANY)    ('*LOAD DEST ''nil) (NEGMAC L1 A1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1)) 

  ( ANY            ('*USES DEST)       ('*LOAD T1    A1) 
                   ('*LOAD DEST ''nil) (NEGMAC L1 T1 A2) 
                   ('*LOAD DEST ''t)   ('*LBL  L1))))
  


(PUT 'TVPAT1 'PATTERN '(nil             %  ATOM, PAIRP,  etc FN
    NIL
    ((DREGDEST) ('*LOAD T1   ''nil) (NEGMAC L1 A1 P2) 
                ('*LOAD T1   ''t)   ('*LBL  L1) ('*RETURN T1) 
		('*SET DEST (FN A1)) ('*DESTROY A1) ) 
    ((DREGNP)   ('*LOAD DEST ''nil) (NEGMAC L1 A1 P2) 
                ('*LOAD DEST ''t)   ('*LBL  L1) 
		('*SET DEST (FN A1)) ('*DESTROY A1) ) 
    ( ANY       ('*USES DEST)       ('*LOAD TD1  A1) 
                ('*LOAD DEST ''nil) (NEGMAC L1 TD1 P2) 
                ('*LOAD DEST ''t)   ('*LBL  L1) 
		('*SET DEST (FN A1)) ('*DESTROY TD1) )
    ))
    


(PUT 'TSTPAT 'PATTERN '(nil             %  EQ, NE, WLEQ  etc   TST
    
    &FIXREGTEST 
    ((AFRAME ZERO) ('*LOAD TD1 A1) (MAC DEST TD1 A2)) 	%  TOO BAD   scs
    ((ZERO AFRAME) ('*LOAD TD2 A2) (MAC DEST A1 TD2)) 	%  TOO BAD   scs
    ((ANY    AREG)                 (MAC DEST A1  A2)) 
    ((AREG   ANY)                  (MAC DEST A1  A2)) 
    ((ANY    ZERO)                 (MAC DEST A1  A2)) 
    ((ZERO   ANY)                  (MAC DEST A1  A2)) 
    ((ANY  SMALL)  ('*LOAD TD2 A2) (MAC DEST A1 TD2)) 
    ((SMALL  ANY)  ('*LOAD TD1 A1) (MAC DEST TD1 A2)) 
    ((ANY    DREG)                 (MAC DEST A1  A2)) 
    ((DREG   ANY)                  (MAC DEST A1  A2)) 
    ((ANY  CONST)                  (MAC DEST A1  A2)) 
    ((CONST  ANY)                  (MAC DEST A1  A2)) 
    ( ANY          ('*LOAD T1 A1)  (MAC DEST T1  A2))))
    


(PUT 'TSTPAT2 'PATTERN '(nil            %  ATOM, PAIRP  etc  TST
     &SETREGS1 
     ((DREGNP)                  (MAC DEST  A1 P2) ('*DESTROY A1)) 
     ( ANY      ('*LOAD TD1 A1) (MAC DEST TD1 P2) ('*DESTROY TD1))))
     


(de &ALTAREG (ARGS) 
    (cond 
     (ARGS 
      % first A register   scs
      (cons (cond 
	     ((DREGP (CAR ARGS))
	      (prog (D I R ACTUALVAL)
		    (setq ACTUALVAL (&REMREGS (setq D (CAR ARGS))))
		    (setq I 9)
		    (RETURN (OR 
			     (WHILE (leq I MAXNARGS&)
				    (progn
				      (cond 
				       ((&SAMEVAL (setq R (&MKREG I))
						  D ACTUALVAL) 
					(RETURN R)))
				      (setq I (plus I 1))))
			     D))))
	      (t 
	       (CAR ARGS)))
	    (&ALTAREG (CDR ARGS))))))

(de &ALTDEST (ARGS) 
 (cond 
  ((OR (&USESDEST (CAR ARGS))                       % scs
       (&USESDEST (CADR ARGS)))
  (setq DEST& (&GETTEMP 'TD nil nil)))) 
 ARGS)




