%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           IBMXA-ASM.SL
% Title:          IBM/370 XA specific information for LAP-TO-ASM
% Author:         Winfried Neun
% Created:        5 Mar 1991
% Package:        Compiler
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
% Dec 12 1988 (Winfried Neun)
%       adapted symfnc preset to new calling scheme
%       removing things that are no more needed
% Aug 22 1988 (padget and burdorf)
%       added function to handle the sun 4 os 4 assembler directive 
%       wierdness which requires the syntax to be 3+(3>>5) instead of
%       the standard (3>>5)+3
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (progn (load if!-system common)
        (off usermode)))

(fluid '(codefilenameformat!* datafilenameformat!* inputsymfile!* 
         outputsymfile!* commentformat!* labelformat!* 
         externaldeclarationformat!* exporteddeclarationformat!* 
         fullwordformat!* doublefloatformat!* reservezeroblockformat!* 
         reservedatablockformat!* definedfunctioncellformat!* 
         undefinedfunctioncellinstructions!* mainentrypointname!* 
         codeout!* dataout!* !*lower asmopenparen!* asmcloseparen!* 
         numericregisternames!* pathin!*))
%%

(fluid '(*declarebeforeuse))
(setq *declarebeforeuse t)

(setq codefilenameformat!* "%w.s")

(setq datafilenameformat!* "d%w.s")

(setq inputsymfile!* "ibmxa.sym")

(setq outputsymfile!* "ibmxa.sym")

(setq mainentrypointname!* (intern "_psl_main"))

(setq commentformat!* "* %p %n")

(setq labelformat!* "%w   EQU  *%n")

(setq exporteddeclarationformat!* "     ENTRY  %w%n")
(setq externaldeclarationformat!* "     EXTRN  %w%n")

(setq fullwordformat!* "        DC A(%e)%n")

(setq reservezeroblockformat!* "%w      DC %eF'0'%n")

(setq reservedatablockformat!* "        DS %eF%n")

(setq definedfunctioncellformat!* " DC  A(%w)%n")

(setq undefinedfunctioncellinstructions!* '((FULLWORD undefinedfunction)))
 
%% From lap-to-asm: 

(de initsymfnc1 (x)
  (prog (ep)
	(setq undefinedfunctioncellinstructions!*
		 `((Fullword ,(get 'undefinedfunction 'entrypoint))))
        (setq ep (get x 'entrypoint))
        (if (null ep)
          (dataprintundefinedfunctioncell (id2int x))
          (dataprintdefinedfunctioncell ep x))))

(de dataprintundefinedfunctioncell (idn)
  (prog (oldout)
        (setq oldout (wrs dataout*))
%        (foreach x in undefinedfunctioncellinstructions* do
%                 (asmoutlap1 x))
        (asmoutlap1 (car undefinedfunctioncellinstructions*))
        (wrs oldout)))

(de dataprintdefinedfunctioncell (x idn)
  (datadeclareexternal x)
     % dont disturb undefined function and interpreter
  (if (memq x '(COMPILEDCALLINGINTERPRETED undefinedfunction))
   (dataprintf " ba %w%n nop%n" x)
   (dataprintf definedfunctioncellformat* x idn)))

(fluid '(&links-present& locallabels*))

(de asmoutlap (u)
  (prog (locallabels* oldout)
        (setf &links-present& (or (look-for-links u)
                                  (assoc '*Id-apply u)
                                  (assoc '*call u)))

	(setq u (append u '(("   DS  0F"))))
        (setq u (pass1lap (&fillframeholes u)))
        (setq u (lapoptreformjumps u)) % lapopt pass
        % Expand cmacros, quoted expressions
        (codeblockheader)
        (setq oldout (wrs codeout*))
        (foreach x in u do (asmoutlap1 x))
        (wrs oldout)
        (codeblocktrailer)))

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(setq asmopenparen!* '!()

(setq asmcloseparen!* '!))

(deflist '((remainder !%) (land !&) (lor !|) (lxor !^) (lsh !<!<)) 
         'binaryasmop)

(put 'lnot 'unaryasmop '!!)

%(put 'mkitem 'asmexpressionformat "(%e<<27)+%e")
(fluid '(*mkitemformat))           %modified for sun 4 os 4
(setq *mkitemformat "      DC  AL1(%e)%n     DC   AL3(%e)") % ZIB

(put 'mkitem 'asmexpressionfunction 'asm-mkitem)

(de asm-mkitem (mkitem-tag-inf)
  (printf
    *mkitemformat
    `(printexpression ,(times2 8 (cadr mkitem-tag-inf)))
    `(printexpression ,(caddr mkitem-tag-inf))))

(global '(*writingasmfile))

(de codefileheader ()

  (setq *writingasmfile t)      %% Somebody must set this flag. but who?

  (codeprintf "CODE   CSECT%n"))

(de datafileheader ()
  (dataprintf "DATA   CSECT%n"))

(de codefiletrailer ()
  (codeprintf "    END %n"))

(de datafiletrailer ()
  )

(de codeblockheader ()
    NIL)

(de codeblocktrailer ()
  )

(de dataalignfullword ()
  (dataprintf " DS  0F%n"))

(de printstring (s)
  (prog (n)
        (setq n (size s))
        (for (from i 0 n 1) (do (printbyte (indx s i))))
        (printbyte 0)
        (printf "     DS  0F%n")))

(de printbytelist (l)
  (foreach x in l do (printbyte x)))

(de printbyte (x)
  (printf "     DC AL1(")
  (printexpression x)
  (printf ")")
  (terpri))

(de printopcode (x)
  (prog (!*lower)
        (prin2 x)))

(de specialactionformainentrypoint ()
  (codeprintf "_psl_main  DS 0F%n"))

(de asmsymbolp (x)
  (ibmxasymbolp (if (idp x)
                 (id2string x)
                 x)))

(de ibmxasymbolp (x)
  (prog (n c i)
        (setq n (size x))
	(when (wgreaterp n 7) (return nil))
        (setq c (indx x 0))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i 1)
   loop
        (when (greaterp i n)
          (return t))
        (setq c (indx x i))
        (unless (or (and (geq c (char a)) (leq c (char z))) 
                    (and (geq c (char !0)) (leq c (char !9))) 
                    (and (geq c (char !a)) (leq c (char !z))))
          (return nil))
        (setq i (plus i 1))
        (go loop)))

(de printnumericoperand (x)
  (prin2 x))

(de operandprintdeferred (x)
  (setq x (cadr x))
  (if (regp x)
    (progn (prin2 '!(!,)
           (printoperand x)
           (prin2 '!)))
    (error 0 (list "In sparc, differed operand can't be used like this:" x))))

(put 'deferred 'operandprintfunction 'operandprintdeferred)

(de operandprintdisplacement (x)
  (setq x (cdr x))
  (printexpression (cadr x))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!)))

(put 'displacement 'operandprintfunction 'operandprintdisplacement)

(de operandprintxdisplacement (x)
  (setq x (cdr x))
  (if (pairp (cadr x))
      (printexpression (times2 4 (findidnumber (cadr (cadr x)))))
      (printexpression (times2 4 (findidnumber (cadr x)))))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!)))

(put 'xdisplacement 'operandprintfunction 'operandprintxdisplacement)

(de operandprintxxdisplacement (x)
  (setq x (cdr x))
  (if (pairp (cadr x))
      (printexpression (findidnumber (cadr (cadr x))))
      (printexpression (findidnumber (cadr x))))
  (prin2 '!( )
  (printoperand (car x))
  (prin2 '!)))

(put 'xxdisplacement 'operandprintfunction 'operandprintxxdisplacement)

(de operandprintidloc (x)
  (setq x (cadr x))
  (codeprintf "    DC   V(%w)%n" (findentrypoint x) )
  )

(put 'idloc 'asmpseudoop 'operandprintidloc)

(de operandprintfullindexed (x)
  (setq x (cdr x))
  (printoperand (caddr x))
  (prin2 '!( )
  (printoperand (cadr x))
  (prin2 '!, )
  (printoperand (car x))
  (prin2 '!) ))

(put 'doubleindexed 'operandprintfunction 'operandprintfullindexed)

(de operandprintimmediate (x)
  (printexpression (cadr x)))

(put 'immediate 'operandprintfunction 'operandprintimmediate)

(de operandprintforeignentry (x)
  (prog (!*lower symnamx)
        (prin2 (cadr x))))

(put 'foreignentry 'operandprintfunction 'operandprintforeignentry)

(de mcprint (x)
  % Echo of MC's                                                           
  (codeprintf "* %p %n" x))

(de instructionprint (x)
  (codeprintf "*         %p %n" x))

(de !*cerror (x)
  (prog (i)
        (setq i (wrs nil))
        (printf "%n *** CERROR: %r %n " x)
        (wrs i)
        (return (list (list 'cerror x)))))

(put 'cerror 'asmpseudoop 'printcomment)

(defcmacro !*cerror)
%%
%%      (word label)
%%
(put 'word 'asmpseudoop 'print-word)

(de print-word (x)
  (prin2 " DS  F(")
  (printoperand (cadr x))
  (prin2 ")")
  (terpri))

(put '*entry 'asmpseudoop 'asmprintentry)
(fluid '(*mainfound))
(de asmprintentry (x)
  (prog (y)
        (printcomment x)
        (setq x (cadr x))
        (setq y (findentrypoint x))
        (unless (flagp x 'internalfunction)
          (findidnumber x))
        (if (eq x mainentrypointname*)
          (progn (setq *mainfound t)
                 (specialactionformainentrypoint))
          (codedeclareexporteduse y))))

(de codedeclareexporteduse (y)
    (codedeclareexported y)
    (printlabel y)
    )

(de findentrypoint (x)
  (cond ((get x 'entrypoint) (get x 'entrypoint))
        ((and (asmsymbolp x)
              (not (get x 'symbol))
              (not (flagp x 'foreignfunction)))
         (put x 'entrypoint x) 
         x)
        (t 
         (let ((name (stringgensym)))
           (put x 'entrypoint name)
           name))
        ))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = "))

%%
%%      define pseude operation
%%
(de printdefine (x)
  (prin2 (cadr x))
  (prin2 " = ")
  (prin2 (caddr x))
  (terpri))

(put 'define 'asmpseudoop 'printdefine)

(de print-register-or-immediate(name)
  (cond ((regp name) (printoperand name))
        (t (printoperand name))))

(de print-immediate (val)
  (cond ((atom val)
         (printoperand val))
        ((eq (car val) 'WARRAY)
         (prin2 (cadr val)))
        (t
         (error 0 (list "Unknown constant expression:" val)))))

(de print-register-name(name)
  (setq name (cadr name))
  (cond ((and (numberp name) (lessp name 6) (greaterp name 0))
			 (prin2 (nth '("4" "5" "6" "7" "8") name)))
        ((get name '*register-print-name)
         (prin2 (get name '*register-print-name)))
        (t (printoperand name))))       %% if reg-name is not a register, it
                                        %% must be a constant.

%%
%%
%%      This function is originally defined in lap-to-asm.sl.
%%
(de printoperand (x)
  (cond ((stringp x) (prin2 x))
        ((numberp x) (printnumericoperand x))
        ((regp x) (print-register-name x))
        ((idp x) (prin2 (findlabel x)))
        ((regp (car x)) (print-src-operand x))
        (t (prog (hd fn)
                 (setq hd (car x))
                 (cond ((setq fn (get hd 'operandprintfunction)) 
                        (apply fn (list x)))
                       ((and (setq fn (getd hd)) (equal (car fn) 'macro)) 
                        (printoperand (apply (cdr fn) (list x))))
                       ((setq fn (wconstevaluable x)) (printoperand fn))
                       (t (printexpression x)))))))
%
%       print operand by [rs1+rs2] format.
%
(de print-src-operand (src)
  (prin2 "[")
  (printoperand (car src))
  (and (cdr src)
       (print-secondary-source (cadr src)))
  (prin2 "]")
  )

(de print-secondary-source (src)
  (cond ((and (pairp src)
              (eq (car src) '-))
         (cond ((numberp (cadr src))
                (cond ((minusp (cadr src))
                       (prin2 "+")
                       (prin2 (minus (cadr src))))
                      (t
                       (prin2 "-")
                       (prin2 (cadr src)))))
               (t
                (prin2 "-")
                (printoperand (cadr src)))))

        ((numberp src)
         (cond ((minusp src)
                (prin2 "-")
                (prin2 (minus src)))
               (t
                (prin2 "+")
                (printoperand src))))
        (t
         (prin2 "+")
         (printoperand src))))

(deflist
  '((r0  "0")(r1  "1")(r2  "2")(r3  "3")(r4  "4") (r5  "5")(r6  "6")
	(r7  "7")(r8  "8")(r9  "9") (r10 "10")(r11 "11")(r12 "12")
	(r13 "13")(r14 "14") (r15 "15") (t1 "10") (t2 "11") (st "13"))
  '*register-print-name)

(put 'nil '*register-print-name "9")

(de asmoutlap1 (x)
  (prog (fn)
        (return (cond ((stringp x) (printlabel x))
                      ((atom x) (printlabel (findlocallabel x)))
                      ((setq fn (get (car x) 'asmpseudoop))
                       (apply fn (list x)))
                      (t
                       % instruction output form is:
   % "space" <opcode> [ "space" <operand> { "comma" <operand> } ] "newline "

      (progn (prin2 " " ) 
             (printopcode (car x))
             (setq x (cdr x))
             (unless (null x)
                     (prin2 " ")
                     (printoperand (car x))
                     (foreach u in (cdr x) do
                              (progn (prin2 '!,)
                                     % COMMA
                                     (printoperand u))))
             (prin2 !$eol!$)))))))

(de datareservefunctioncellblock (x)
 (rplaca printexpressionformpointer* x)
 (dataprintf reservedatablockformat* printexpressionform*))


(de datareserveblock (x)
 (rplaca printexpressionformpointer* x)
 (dataprintf reservedatablockformat* printexpressionform*))

(de asmpseudoprintfullword (x)
  (if (eqcar (cadr x) 'mkitem)
     (if (eqcar (caddr (cadr x)) 'idloc)
         (printf  "      DC  AL1(%w)%n     DC   AL3(%w)%n"
	(times2 8 (cadr (cadr x))) (findidnumber (cadr (caddr (cadr x)))))
	 (printf  "      DC  AL1(%w)%n     DC   AL3(%w)%n" 
			 (times2 8 (cadr (cadr x))) (caddr (cadr x))))
    (foreach y in (cdr x) do (printfullword y))))

(de initializesymnam (maxsymbol)
   
    (setq strifile (open "smain.s" 'output))

    (dataprintgloballabel (findgloballabel 'symnam))
    (for (from i 0 128 1) (do
        (dataprintf "   DC AL1(32)%n   DC  AL3(%w)%n" 
		(caddr (compileconstantxx (id2string (int2id i)))))))
    (foreach idname in (car orderedidlist*) do
        (dataprintf "   DC AL1(32)%n   DC  AL3(%w)%n"
                (caddr (compileconstantxx (id2string idname)))))
    (channelprintf strifile "     END ")
    (close strifile)
)

(de compileconstantxx (x)
  (prog (oldout* cdrx)  
   (setq oldout* (wrs dataout*))
  (setq x (buildconstant x))
  (if (null (cdr x))
    (car x)
    (progn (wrs strifile) (setq cdrx (cdr x))
           (foreach xx in cdrx do (asmoutlap1 xx))))
  (wrs oldout*)
  (return (car x))))
