%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXK:EXTERNALS.SL
% Description:  Definitions of foreign language functions
% Author:       Brian Beach, Hewlett-Packard CRC
% Created:      19-Apr-84
% Modified:     15-Feb-85 13:25:43
% Mode:         Lisp
% Package:      
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 29-May-87 (Harold Carr & Leigh Stoller)
%  Added external definition for setenv.
% 05-May-87 (Leigh Stoller)
%  Added external definitions for fast-math.sl. They are in float.c
% 13-Apr-87 (Leigh Stoller & Harold Carr)
%  Added external definition of allocatemorebps, defined in bpsheap.c.
% 26-Sep-86 (Leigh Stoller)
%  Added external entry for exit which is needed on the sun to pass an exit
%  code back to unix.
% 01-Sep-86 (Leigh Stoller)
%  Added external-system to do a standard system call from C.
% 19-Aug-86 (Leigh Stoller)
%  Added entries for get_file_status, sigrelse, and sigset.
% 15-Feb-85 13:25:18 (Brian Beach)
%  Commented out a couple of functions for Pisces.
% 01-Feb-85 (Bill Williams)
%  Add external entries for Lisp terminal emulator.
% 24-Jan-85 (Vicki O'Day)
%  Added sleep entry.
% 7-Jan-85 (Vicki O'Day)
%  Removed nmodekeys entry.
% 17-Dec-84 (Vicki O'Day)
%  Added setlinebuf entry.
%  Added syscall entry, for access to HP-UX system calls.  The syntax
%  of this command is (syscall <system call #> <arg1> <arg2> ...).
%  Currently, a limit of four arguments to the system call is imposed,
%  since no system calls that we can use have more than this.  The
%  various forms of execl do, but since we can't use fork now, we don't
%  need this.  When we can use fork, the number of arguments allowed
%  by syscall should be increased.
% 7-Dec-84 (Vicki O'Day)
%  Added nmodekeys entry, so 9836-users can call it.
% 27-Sep-84 19:57:38 (Dave Matthews)
%  Added an external function declaration for the new alterheapsize function.
% 17-Jul-84 22:51:12 (RAM)
%  Changed chdir to unixcd and time to external_time.
%  Added external_stat, link, and unlink.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime
 (setf *foreign-functions* nil)
 )

(compiletime 
 (defmacro external-function (name args)
     (setf *foreign-functions* (cons name *foreign-functions*))
     (flag1 name 'foreignfunction)
     `(de ,name ,args
        (,name ,@args)
        )))

(off r2i)

% Defined in echo.c
%
(external-function echoon               ())
(external-function echooff              ())
(external-function external_charsininputbuffer    (chan))
(external-function flushstdoutputbuffer ())
(external-function external_user_homedir_string    ())
(external-function external_anyuser_homedir_string (user))


% Defined in bpsheap.c
%
(external-function alterheapsize        (integer))
(external-function allocatemorebps	(amount))

% Defined in file-status.c
%
(external-function get_file_status      (filename block flag))


% Defined in os-hooks.c
%
(external-function os_startup_hook      (pargc pargv))
(external-function os_cleanup_hook      (retv))
 

% Defined in pslextras.c
%
(external-function external_alarm      (sec))
(external-function external_ualarm     (usec repeat))
(external-function external_time       (buffer))
(external-function external_timc       (buffer))
(external-function external_stat       (path buf))
(external-function external_link       (path1 path2))
(external-function external_unlink     (path))
(external-function external_strlen     (strptr))
(external-function external_setenv     (varstring valstring))
(external-function external_getenv     (envstring))
(external-function uxfloat             (buffer integer))
(external-function uxfix               (buffer))
(external-function uxassign            (arg1-buffer arg2-buffere))
(external-function uxplus2             (result-buffer arg1-buffer arg2-buffer))
(external-function uxdifference        (result-buffer arg1-buffer arg2-buffer))
(external-function uxtimes2            (result-buffer arg1-buffer arg2-buffer))
(external-function uxquotient          (result-buffer arg1-buffer arg2-buffer))
(external-function uxgreaterp          (arg1-buffer arg2-buffer tee nill))
(external-function uxlessp             (arg1-buffer arg2-buffer tee nill))
(external-function uxwritefloat        (buffer floatptr convstr))
(external-function uxdoubletofloat     (x y))
(external-function uxfloattodouble     (y y))
(external-function uxsin	       (r x))
(external-function uxcos	       (r x))
(external-function uxtan	       (r x))
(external-function uxasin	       (r x))
(external-function uxacos	       (r x))
(external-function uxatan	       (r x))
(external-function uxsqrt	       (r x))
(external-function uxexp	       (r x))
(external-function uxlog	       (r x))
(external-function uxatan2	       (r y x))
(external-function cremainder (x y))
(external-function cquotient  (x y))


% Defined in pwd-fn.c
%
(external-function external_pwd         ())
(external-function datetag())           % returns date of bpsl build


% Defined in sigs.c
%
(external-function psl_sigset           (signame handler))
(external-function sigrelse             (signame handler))


% Defined In unexec.c
%
(external-function unexec               (newname aname dstart bstart))


% Defined in unix-io.c
%
(external-function unixputc             (ch))
(external-function unixputs             (str))
(external-function unixputn             (num))
(external-function unixcleario          ())
(external-function expand_file_name     (str))
(external-function unixopen             (name mode))
(external-function unixgetw             (file))
(external-function unixcd               (dir))
(external-function ctime                (buffer))
(external-function external_system      (command))


% Defined In the C Library
%
(external-function external_exit        (status))
(external-function fopen                (name mode))
(external-function fclose               (fp))
(external-function fread                (buf size count fp))
(external-function fputc                (ch fp))
(external-function fgetc                (fp))
(external-function fgets                (buf count fp))
(external-function fwrite               (str strlen count fp))
(external-function fflush               (fp))
(external-function fseek                (fp offset ptrname))
(external-function clearerr             (fp))
(external-function getw                 (fp))
(external-function putw                 (w fp))
(external-function signal               (signame handler))
(external-function sleep		(sec))
(external-function setjmp               (env))  % added for sparc
(external-function longjmp              (env val)) % to process interrupts.
(external-function perror               (strbase)) % for errno evalualtion


%% See $pxnk/sys-io.sl for the call to this. In SYS V, it may have to be
%% moved into the microkernel since it is not directly supported. Under
%% SYS V, buffering can only be changed before any writes on the stream,
%% which means it must be called before setupbpsandheap. Also, under BSD,
%% setlinebuf is a system call, while under SYS V, it must be written using
%% setbuf and setvbuf. See the Bobcat version of setlinebuf.c in $pb.
%(external-function setlinebuf           (iobu))

(external-function cflush (ad len inv)) % cache flush
(external-function getpid ())
%%socketinterface

(external-function unixsocketopen (name number))
(external-function getsocket (fd name number))
(external-function writesocket (fd name number))
(external-function unixclosesocket (number))

%% fork and wait

%(external-function fork())
%(external-function wait(ptr))

%% popen pclose

(external-function popen (cmd mode))
(external-function pclose (stream))

%% shared memory , semaphores

(external-function shmctl (shmid cmd buf))
(external-function shmget (key size shmflg))
(external-function shmat  (shmid shmaddr  shmflag))
(external-function shmdt (shmaddr))

(external-function semctl (semid semnum cmd arg))
(external-function semget (key nsems semflg))
(external-function semop  (semid sembuf nsops))

% pvm

(external-functionXTvm_addhosts (hosts nhost infos))
% (external-function pvm_advise (a1 a2 a3))
(external-function pvm_barrier  (group count))
(external-function pvm_bcast    (group msgtag))
(external-function pvm_bufinfo  (bufid bytes msgtag tid))
(external-function pvm_config   (nhost narch hostp))
(external-function pvm_delhosts (hosts nhost infos))
(external-function pvm_exit     ())
(external-function pvm_freebuf  (bufid))
(external-function pvm_getinst  (group tid))
(external-function pvm_getrbuf  ())
(external-function pvm_getsbuf  ())
(external-function pvm_gettid   (group inum))
(external-function pvm_gsize    (group))
(external-function pvm_initsend (encoding))
(external-function pvm_joingroup (group))
(external-function pvm_kill     (tid))
(external-function pvm_lvgroup  (group))
(external-function pvm_mcast    (tids ntask msgtag))
(external-function pvm_mkbuf    (encoding))
(external-function pvm_mstat    (host))
(external-function pvm_mytid    ())
(external-function pvm_notify   (what msgtag ntask tids))
(external-function pvm_nrecv    (tid msgtag))
(external-function pvm_pkbyte   (pointer nitem stride))
(external-function pvm_pkcplx   (pointer nitem stride) )
(external-function pvm_pkdcplx  (pointer nitem stride) )
(external-function pvm_pkdouble (pointer nitem stride) )
(external-function pvm_pkfloat  (pointer nitem stride) )
(external-function pvm_pkint    (pointer nitem stride) )
(external-function pvm_pklong   (pointer nitem stride) )
(external-function pvm_pkshort  (pointer nitem stride) )
(external-function pvm_pkstr    (pointer) )
(external-function pvm_parent   ())
(external-function pvm_perror   (msg))
(external-function pvm_probe    (tid msgtag))
(external-function pvm_pstat    (tid))
(external-function pvm_recv     (tid msgtag))
(external-function pvm_recvf    (new))
(external-function pvm_send     (tid msgtag))
(external-function pvm_sendsig  (tid signum))
% (external-function pvm_serror (set))
(external-function pvm_setrbuf  (bufid))
(external-function pvm_setsbuf  (bufid))
(external-function pvm_spawn    (task argv flag where ntask tids))
(external-function pvm_tasks    (where ntask taskp))
(external-function pvm_upkbyte  (pointer nitem stride))
(external-function pvm_upkcplx  (pointer nitem stride) )
(external-function pvm_upkdcplx (pointer nitem stride) )
(external-function pvm_upkdouble (pointer nitem stride) )
(external-function pvm_upkfloat (pointer nitem stride) )
(external-function pvm_upkint   (pointer nitem stride) )
(external-function pvm_upklong  (pointer nitem stride) )
(external-function pvm_upkshort (pointer nitem stride) )
(external-function pvm_upkstr   (pointer) )

(on r2i)

(compiletime
 (for (in name *foreign-functions*)
      (do (remflag1 name 'foreignfunction))
      ))

(compiletime (flag '(profil) 'foreignfunction))
(on r2i)

