%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           PXK:MAIN-START.SL
% Title:          First routine called on startup
% Author:         Brian Beach, Hewlett-Packard CRC
% Created:        16-Feb-84
% Modified:       May 16 1985
% Status:         Experimental
% Mode:           Lisp
% Package:        Kernel
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime    %% Move this hack somewhere else later.
 (defmacro exported-global (list)
   (for (in id (second list))   % list is quoted
        (do (prog (*lower)
         (setq *lower t)
         (flag1 id 'exportedsymbol)
         (put id 'symbol (bldmsg "_%w" id))
         (put id 'vartype 'global)
         (findidnumber id)
         )))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro internal-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (flag1 id 'internalsymbol)
         (put id 'symbol id)
         (put id 'vartype 'global)
         (findidnumber id)
         ))
   ))

(compiletime    %% Move this hack somewhere else later.
 (defmacro system-global (list)
   (for (in id (second list))   % list is quoted
        (do 
         (put id 'vartype 'global)
         (findidnumber id)
         ))
   ))

(on fast-integers)

(compiletime (load sys-consts))

(declare-kernel-structure 
 (stack             #.(compiler-constant 'initial-stacksize))
 (argumentblock     #.(compiler-constant 'maxargs))
 (tokenbuffer       #.(compiler-constant 'maxtokensize))
 (bndstk            #.(compiler-constant 'initial-bndstksize))
 (catchstack        #.(times2 (compiler-constant 'catchstacksize) 4))
 (hashtable         #.(quotient (compiler-constant 'hash-table-size) 2))
 (onewordbuffer     1)
 (saveargc          1)
 (saveargv          1)
 (datebuffer        1)
 )

(exported-global
 '(bndstkptr   bndstkupperbound   gcarraylowerbound   gcarrayupperbound
   heaplast   oldheaplast
   heaplowerbound  heaptrapbound   oldheaptrapbound   heapupperbound
   _infbitlength_   lastbps   bpslowerbound   mainstartinitialize
   nextbps   oldheapupperbound   oldheaplowerbound
   stacklowerbound   stackupperbound   saveda7
   unixstdin   unixstdout   unixstderr   unixnull   unixeof   unixtty
   ualarm 
   ))

(internal-global  '(   symnam  symval  symget symfnc  symprp   ))

(system-global  '(symnam symval symprp  symfnc symget
                  %  stack
                  catchstackptr   catchstack
                  bndstklowerbound   bndstkupperbound
                  bndstkptr   bndstk
                  heap
                  nextsymbol    %% Had SYMBOL as its assembler symbol.  why????
                    % Add these so they are always availbale to compiled code.
                  ***MUST-BE-NIL***   ***MUST-BE-NIL-TOO***
                  *fastcar
   ))

(compiletime (setq *syslisp nil))

(de init-pointers()
  (setq catchstackptr catchstack)
  (setq bndstklowerbound bndstk)
  (setq bndstkupperbound (loc (wgetv bndstk (wdifference bndstksize 1))))
  (setq bndstkptr bndstk)
  (setq heap heaplowerbound)
  )

(de init-fluids-old ()
        % Until problem of labels not appearing is fixed...
  (setq gcarraylowerbound nil)
  (setq gcarrayupperbound nil)
  (setq heaplast nil)
  (setq oldheaplast nil)
  (setq heaptrapbound nil)
  (setq oldheaptrapbound nil)
  (setq heapupperbound nil)
  (setq _infbitlength_ nil)
  (setq lastbps nil)
  (setq bpslowerbound nil)
  (setq mainstartinitialize nil)
  (setq nextbps nil)
  (setq oldheapupperbound nil)
  (setq oldheaplowerbound nil)
  (setq stackupperbound nil)
  (setq unixstdin nil)
  (setq unixstdout nil)
  (setq unixstderr nil)
  (setq unixeof nil)
  (setq unixnull nil)
  (setq unixtty nil)
  (setq ualarm nil)
  (setq ***MUST-BE-NIL*** NIL)
  (setq ***MUST-BE-NIL-TOO*** NIL)
  (setq *fastcar nil))

(compiletime (setq *syslisp t))

(de init-fluids ()
        % Until problem of labels not appearing is fixed...
  (init-fluids-old) % do you know the difference?
  (setq gcarraylowerbound nil)
  (setq gcarrayupperbound nil)
  (setq heaplast nil)
  (setq oldheaplast nil)
  (setq heaptrapbound nil)
  (setq oldheaptrapbound nil)
  (setq heapupperbound nil)
  (setq _infbitlength_ nil)
  (setq lastbps nil)
  (setq bpslowerbound nil)
  (setq mainstartinitialize nil)
  (setq nextbps nil)
  (setq oldheapupperbound nil)
  (setq oldheaplowerbound nil)
  (setq stackupperbound nil)
  (setq unixstdin nil)
  (setq unixstdout nil)
  (setq unixstderr nil)
  (setq unixnull nil)
  (setq unixeof nil)
  (setq unixtty nil)
  (setq ualarm nil)
  (setq ***MUST-BE-NIL*** NIL)
  (setq ***MUST-BE-NIL-TOO*** NIL)
  (setq *fastcar nil)
)

(compiletime (on mainfound))

(de drucken (txt nbr)
    (console-print-string txt)
    (unixputn nbr)
    (console-newline))

(lap '(("        .globl _psl_main ")
       ("_psl_main: ")
       (*entry _!p!s!l_!m!a!i!n expr 0)
       (*alloc 1) 
       (*move (reg st) ($fluid saveda7))                 % save A7
       (*move (reg st) ($fluid stacklowerbound))
 
                           % save start parameters
       (*move (indexed (reg a6) 0) ($fluid argc)) 
       (*move (indexed (reg a6) 4) ($fluid argv))
 
       (*move   (reg st)      ($fluid stackupperbound))
 
              %  Do OS specific initializations (uses argc and argv)
       (*move ($fluid argc) (reg 1))
       (*move ($fluid argv) (reg 2))
       (*move infbitlength ($fluid _infbitlength_))
       (*link os_startup_hook expr 2)
 
       (*call init-pointers)
 
                   % handcoded access to the fluid cell NIL
       (*move ($fluid symval) (reg a1))
       (*move (indexed (reg a1)(wconst (times 128 addressingunitsperitem)))
                (reg nil))                    % initialize NIL reg
 
       (*call pre-main)                                 % call PSL
 
panic-exit                      % need to do UNIX cleanup after
                                % a fatal error, so jump here
       (*move (wconst 0) (reg 1))
       (*linke 0 exit-with-status expr 1)
       (*exit 0)
 
       (*entry exit-with-status expr 1)
  %    (*push (reg 1))
  %    (*link os_cleanup_hook expr 0)
  %    (*pop (reg 1))
       (*move ($fluid saveda7) (reg st))
       (*link external_exit expr 1)
       (*exit 0)
       ))

(de pre-main ()
  (unixcleario)
  (initialize-symbol-table)
  (initcode)
  (console-print-string "About to load LOAD-PSL")
  (console-newline)
  (faslin "load-psl.b")
  (loader-main))

% hook for catching a running program with adb ":b" feature
 
(lap '((*entry adb expr 1)
       (*move (indexed (reg fp) 0) (reg 2)) % display return address
       (*exit 0)))

(off fast-integers)
