%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PK:FASLIN.SL 
% Title:        Loading of binary format files. 
% Author:       E. Benson 
% Created:      ??? 
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  PL:FASL-DECLS.B 
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 14-Jun-88 (Tsuyoshi Yamamoto)
%  Added new relocation types (relocate-movex,relocate-call) and long word
%  alignment for halfword access.
% 10-Jan-84 (Brian Beach)
%  Commented-in checking of fasl-magic-number.
% 13-Nov-84 14:13:11 (Brian Beach)
%  Moved FASLIN-INTERN to INTERN.SL.
% 10-May-84 14:35:05 (Brian Beach)
%  Changed fasl-magic-number to faslin-magic (which is defined now in SYS-CONSTS.)
% 01-Dec-83 14:44:33 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (load fasl-decls))

% a lot of horrible hack here, sorry about that

% this does 'real' lower 16 bits (shift by two ) number returning
(compiletime (put 'wandffff 'opencode '((ori (reg t1) (reg zero) 16#ffff)
                                        (and (reg 1) (reg 1) (reg t1))
                                        (sll (reg 1) (reg 1) 2))))

% same with 1st argument shift by -2
(compiletime (put 'wandffff2 'opencode '((ori (reg t1) (reg zero) 16#ffff)
					(srl (reg 1) (reg 1) 2)
                                        (and (reg 1) (reg 1) (reg t1))
                                        (sll (reg 1) (reg 1) 2))))

% this does 'real' upper16 bits
(compiletime (put 'wandffff3 'opencode '((ori (reg t1) (reg zero) 16#ffff)
                                        (nor (reg t1) (reg t1) (reg zero))
                                        (and (reg 1) (reg 1) (reg t1)))))

% wor with second argument shifted by 2
(compiletime (put 'worspec  'opencode '((srl (reg 2) (reg 2) 2)
                                        (or (reg 1) (reg 1) (reg 2)))))

% rightshift by 6 without blanking lower two bits
(compiletime (put 'wshift-6 'opencode '((srl (reg 1) (reg 1) 6))))

% add a 'hard' 7
(compiletime (put 'wplus2-7 'opencode '((addiu (reg 1) (reg 1) 7))))

(on fast-integers)

(fluid '(code-base-hack
	 symfnc
	 symval
	 tokenbuffer
	 argumentblock))

(de faslin (file)
  (prog (fid                  % file pointer 
	 local-id-count       % number of ids in the file
	 local-id-table       % table for mapping local ID numbers to global ID numbers.
	 code-size            % number of words of code
	 code-base            % location of the start of the code
	 init-function-address% Offset into the code of the init function
	 bit-table-size       % number of words in the bit table
	 bit-table            % the bit table
	 Btop
	 )

    % Open the file
    (setf fid (binaryopenread file))

    % Check that the first word is the correct magic number.
    (let ((first-word (binaryread fid)))
      (unless (weq first-word faslin-magic)
	(binaryclose fid)
	(faslin-bad-file file)
	(return nil)
	))

    % Read in the ID table.
    (setf local-id-table (read-id-table fid))

    % Read the code.
    (setf code-size (binaryread fid)) % Size of code segment in words
    (setf code-base (gtbps code-size)) % Allocate space in BPS
    (setq Btop (GtBPS 0))              % pointer to top of alloc. BPS
    (setf init-function-address (wplus2 code-base (binaryread fid)))
    (binaryreadblock fid (loc (wgetv code-base 0)) code-size)

    % Read the bit table
    (setf bit-table-size (binaryread fid))
    (setq bit-table (gtwrds bit-table-size))
    (setq bit-table (mkwrds bit-table))
    (binaryreadblock fid (loc (words-fetch bit-table 0)) bit-table-size)

    % Close the file
    (binaryclose fid)

    % Twiddle the bits.
    (do-relocation code-base code-size bit-table local-id-table)

    % Call the init code
    (let ((temp code-base-hack))  % avoid use of fluid binding
      (setf code-base-hack code-base)
      (addressapply0 init-function-address)
      (setf code-base-hack temp)
      (DelBPS (wplus2 init-function-address 1) Btop)
      )
    ))


(define-constant reloc-movex 1)
(define-constant reloc-call  2)
(define-constant reloc-xidloc 3)

(de do-relocation (code-base code-size bit-table id-table)
  % CODE-AU-SIZE is the size of the code measured in addressing
  % units, rather than words.

  (let ((code-au-size code-size )) %addressingunitsperitem)))
    (for (from i 0 (wdifference code-au-size 1) 1)
     (do 
      (let ((bit-table-entry  (bittable (loc (words-fetch bit-table 0)) 
					(wshift i 2)))
	    (code-location    (wplus2 code-base i ))	memo )
	(case bit-table-entry
	 ((reloc-word) % Mips (and  Sparc) specific escape
	  (case (bittable (loc (words-fetch bit-table 0))
					 (wplus2 (wshift i 2) 1))
	 	((0) (relocate-word code-location code-base id-table))
		((reloc-movex)
		    (relocate-movex code-location code-base id-table))
		((reloc-call)
		    (progn (setq memo (getmem code-location))
		            (setq memo 
		                (wplus2 memo (wshift-6 (wshift code-base 4))))
		                        %clear first bits (see J instr!)
		                (setf (getmem code-location) memo)))
		((reloc-xidloc) 
		    (progn (setq memo (wandffff (getmem code-location)))
		           (when (local-id-number? memo)
		               (setq memo (local-to-global-id memo id-table)))
		           (putmem code-location 
		             (worspec (wandffff3 (getmem code-location))
				(wandffff2 (wtimes2 4 (wplus2 memo -8000))) ))
		      ))))
		     %
	((reloc-inf) (relocate-inf code-location code-base id-table))
	((reloc-right-half)
	       (relocate-right-half code-location code-base id-table))
))))))
		    
		     
(de relocate-word (code-location code-base id-table)
  (let ((reloc-tag (reloc-word-tag (getmem code-location)))
	(reloc-inf (reloc-word-inf (getmem code-location))))
    (cond ((eq 0 (wshift (getmem code-location) -20)) 
			% a naked pointer from movex
	     (setf (getmem code-location)
		(wplus2 code-base (getmem code-location)) ))
	  (t (setf (getmem code-location) % an item
		(compute-relocation reloc-tag reloc-inf code-base id-table))))
      ))

(commentoutcode de relocate-inf  (code-location code-base id-table)
  (let ((reloc-tag (reloc-inf-tag (getmem code-location)))
	(reloc-inf (reloc-inf-inf (getmem code-location))))
%%    (setf (inf (getmem code-location))
   (setf (getmem code-location)
    (wplus2-7
      (compute-relocation reloc-tag reloc-inf code-base id-table)
     )
      )))
(de relocate-inf  (code-location code-base id-table)
     (cond ((idp (getmem code-location)) 
             (setf (getmem code-location)
               (worspec 
                (wshift
                (compute-relocation reloc-id-number
                                    (wshift (getmem code-location) -8)
                                    code-base id-table) 8 ) id-tag )))
           (t (setf (getmem code-location)
	     (compute-relocation
		 reloc-code-offset
                (getmem code-location) %%%(wshift (getmem code-location) 2)
		 code-base id-table)
      ))))

(de relocate-right-half (code-location code-base id-table)
  (let ((temp1 (getmem code-location))
	(temp2 nil))
       (setq temp2 (wandffff temp1))
       (cond ((local-id-number? temp2) 
		(setq temp2 (local-to-global-id temp2 id-table))
		(putmem code-location (worspec (wandffff3 temp1) temp2))))))

(de compute-relocation (reloc-tag reloc-inf code-base id-table)
  (case reloc-tag
    ((reloc-code-offset)  (wplus2 code-base reloc-inf ))
    ((reloc-value-cell) 
     (cond ((local-id-number? reloc-inf) 
	      (loc (symval (local-to-global-id reloc-inf id-table))))
	   (t (loc (symval reloc-inf)))))
    ((reloc-function-cell)
     (progn
      (when (local-id-number? reloc-inf)
	(setq reloc-inf (local-to-global-id reloc-inf id-table)))
      (wplus2 symfnc                      %%% Should be (LOC (SYMFNC xxx)) ???
	      (wtimes2 addressingunitsperfunctioncell reloc-inf))))
    ((reloc-id-number)
     (if (local-id-number? reloc-inf)
       (local-to-global-id reloc-inf id-table)
       reloc-inf
       ))
    ))

(de local-to-global-id (local-id-number id-table)
  (wshift (words-fetch id-table (wdifference local-id-number 2048)) 2))

(de read-id-table (fid)
  % Read in the table of local IDs at the front of the FASL file.
  % Each ID is stored as one word which holds the length, followed 
  % by the appropriate number of words holding the string.

  (let* ((local-id-count  (binaryread fid))
	 (id-table        (mkwrds (gtwrds (wplus2 local-id-count 1)))))
    (for 
     (from i 0 local-id-count)
     (do (setf (wgetv tokenbuffer 0) (binaryread fid)) 
			% word is length of ID name
	 (binaryreadblock fid  (loc (wgetv tokenbuffer 1)) 
		               (strpack (wgetv tokenbuffer 0)))
    (SETF (wgetv tokenbuffer 0) (wshift (wgetv tokenbuffer 0) 8)) %  LT WN
	 (setf (words-fetch id-table i) 
%	   (idinf (faslin-intern (mkstr (loc (wgetv tokenbuffer 0))))))
           (wshift (faslin-intern (mkstr (loc (wgetv tokenbuffer 0)))) -10))
	 ))
    id-table
    ))

(de putentry (name type offset)
  % Called by the initcode.
  % CODE-BASE-HACK is set by FASLIN before the initcode is called.
  (putd name type (mkcode (wplus2 code-base-hack offset) )))

(de faslin-bad-file (name)    % To be redefined
  (console-print-string (kernelstring2string "File is not FASL format"))
  (console-newline) )

(commentoutcode de relocate-movex (code-location code-base id-table)
    (let ((val (getmem (wplus2 code-location -1)))
	  (add (getmem code-location))
          (twentysix 26)
	  (lui nil))
	 (setq lui (wor (wshift 8#17 twentysix)
		        (wor (wand (wshift val -16) 16#ffff)
		                  % pull the register out of next instr
		             (wshift (wand (wshift add -21) 16#1f) 16))))
	 (setq add (wor add (wand val 16#ffff)))
	 % adjust in case of not ori instruction
	 (when (and (eq 1 (wshift (wand val 16#ffff) -15))
		    (not (eq 8#15 (wshift add -26))))
		        (setq lui (wplus2 1 lui)))
	 (setf (getmem (wplus2 code-location -1)) lui)
	 (setf (getmem code-location) add)))

(lap '( (*ENTRY RELOCATE-MOVEX EXPR 3)
  (*ALLOC 4) 
  (*MOVE (REG 1) (FRAME 4)) 
  (*MOVE (QUOTE NIL) (REG 4))
  (*MOVE 26 (REG 3))
  (*MOVE (MEMORY (REG 1) (WCONST 0)) (REG 2))
  (*MOVE (MEMORY (REG 1) (WCONST -1)) (REG 1))
  (*MOVE (REG 1) (FRAME 1))
  (*MOVE  15 (REG 4))
  (sll     (REG 4) 26)
  (*MOVE (REG 1) (REG 5))
 % (*WSHIFT (REG 5) (QUOTE -16))
     (SRL (REG 5) (REG 5) 16)
   (*WAND (REG 5) 65535)
   (*WOR (REG 4) (REG 5)) 
   (*MOVE (REG 2) (REG 1))
 % (*WSHIFT (REG 1) -21)
     (SRL (REG 1) (REG 1) 21)
   (*WAND (REG 1) 31)
   (sll     (REG 1) 16)
   (*WOR (REG 4) (REG 1))
   (*MOVE (REG 4) (FRAME 2))
   (*MOVE  65535 (REG 1))
   (*WAND (REG 1) (FRAME 1))
   (*WOR (REG 2) (REG 1)) 
   (*MOVE (REG 2) (FRAME 3)) 
 % (*WSHIFT (REG 1) (QUOTE -15))
     (SRL (REG 1) (REG 1) 15)
   (*JUMPNOTEQ (LABEL G0003) (REG 1) 1)
   (*MOVE (FRAME 3) (REG 1))
 % (*WSHIFT (REG 1) (QUOTE -26))
      (SRL (REG 1) (REG 1) 26)
   (*JUMPEQ (LABEL G0003) (REG 1) 13)
   (*WPLUS3 (REG 1) 1 (FRAME 2))
   (*MOVE (REG 1) (FRAME 2))
   (*LBL (LABEL G0003))
   (*MOVE (FRAME 2) (MEMORY (FRAME 4) (WCONST -1)))
   (*MOVE (FRAME 3) (REG 1))
   (*MOVE (REG 1) (MEMORY (FRAME 4) (WCONST 0)))
   (*EXIT 4)))

% Maintain alignment on discard too! (alloc.sl)

(de delbps (bottom top)      % returns space to Bps
  (when (weq nextbps top)
    (sysmips 3)  % flush the caches
    % reset to discard init-code, but ensure alignment is retained
    (setq nextbps (if (eq 0 (wand bottom 1)) bottom (wplus2 bottom 1)))))


(de flushcache () (sysmips 3))

(off fast-integers)
