%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:PUTD-GETD.SL 
% Title:        Standard Lisp function defining functions 
% Author:       Eric Benson 
% Created:      18 August 1981 
% Modified:     31-May-84 10:51:14 (Brian Beach) 
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 22-August-1990 (Winfried Neun)
%  TYPE information is put into SYMGET vector (slot 0) to spped up
%  (get fn 'type) in eval
% 01-Oct-88 (Tony Hearn)
%  Replaced single use of flag1/remflag1 by flag/remflag to make module
%  less dependent on non-SL functions.
% 23-May-84 16:16:15 (Mike Creech)
%  Rewrote code-number-of-arguments to deal with functions that were
%  defined to have any number of arguments (-1 in *entry definition).
%  Reformatted PUTD to make more readable.
%  Rewrote remd to be more readable.
%  Reformatted getd to be more readable (and changed u => func).
% 01-Dec-83 14:59:44 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(fluid '(symget &symgetsize&))

(setq &symgetsize& 32)                  % default

(de getd (func)
    % Lookup function definition of func.
    (and
        (idp func)
        (not (funboundp func))
        (cons (or (get func 'type)
                  'expr)
              (if
                 (or
                    (flambdalinkp func)
                    (and (flagp func 'planttrampoline)
                         (get func '*lambdalink)))
                 (get func '*lambdalink)
                 (getfcodepointer func)))))

(de remd (func)

    % Remove function definition of func.

    (let
        ((oldgetd (getd func)))
	(when
	     oldgetd
	     % THEN There was an old definition.
	     (makefunbound func)
	     (remprop func 'type)
	     (when (wgetv SYMGET (idinf func))
		(wputv (inf (wgetv SYMGET (idinf func))) 1 NIL))
			 % Anull slot 0 in SYMGET entry
	     (remprop func '*lambdalink))
	% Return the old definition.
        oldgetd)
    )

(fluid '(*redefmsg % controls printing of redefined
	 *usermode % controls query for redefinition
	 ))

(loadtime
  (progn (setq *usermode nil)  % start in system mode
         (setq *redefmsg t)    % message in PutD
	 ))

(fluid '(*comp % controls automatic compilation
	 promptstring*))

(de code-number-of-arguments (code-pointer)

    % Return the number of arguments within the range 0-maxargs, or
    % "any" if function is setup to have any number of arguments, or
    % NIL if not a code-pointer or too many arguments.

    (when
         (codep code-pointer)
	 % THEN We have a real codepointer.
	 (let
	     ((num-args (!%code-number-of-arguments (codeinf code-pointer))))
	     (cond
	          ((weq num-args -1) "any")
		  ((and (wgeq num-args 0)
			(wleq num-args maxargs)) num-args)
		  (t NIL)
		  )
	     )
	 )
    )
%(de cflush (u) (console-print-string "ein cflush")(console-newline))
% (de errorprintf (u) (console-print-string u))
%(console-print-string (symnam 16#4ed))
%(de printf (u) (console-print-string u))
%(de print (u) (console-print-string u))
%(de prin1 (u) (console-print-string u))
%(de prin2 (u) (console-print-string u))
%(de sysreadrec(a b) (console-print-string "sysreadrec") 0)
%(de read()  (console-print-string "ein Ried") 1723)
(de tak (x y z)
  (if (not (wlessp y x)) z
      (tak (tak (wplus2 x -1) y z)
           (tak (wplus2 y -1) z x)
           (tak (wplus2 z -1) x y)))))))

(de putd (fnname fntype fnexp)

  % Install function definition
  %
  % this differs from the SL Report in 2 ways:
  % - function names flagged LOSE are not defined.
  % - 	"      "   which are already fluid or global are defined anyway,
  % with a warning.
  %

  (cond ((not (idp fnname)) (noniderror fnname 'putd))
        ((not (memq fntype '(expr fexpr macro nexpr)))
         (conterror 1305 "%r is not a legal function type" fntype
                    (putd fnname fntype fnexp)))
        ((flagp fnname 'lose)
         (errorprintf "*** %r has not been defined, because it is flagged LOSE"
                      fnname)
         nil)
        (t (prog (vartype
                  u
		  printredefinedmessage
		  oldin
		  promptstring*
                  queryresponse)
                 (unless
		       (funboundp fnname)
		       % THEN
		       (when
			    *redefmsg
			    % THEN
			    (setq printredefinedmessage t))
		       (when
			    (and *usermode
				 (not (flagp fnname 'user)))
			    % THEN
			    (if
			       (not (yesp (bldmsg
                        "Do you really want to redefine the system function %r?"
                                                  fnname)))
			       % THEN
			       (return nil)
			       % ELSE
			       (flag1 fnname 'user))))

                 (when (and (setq u (getd fnname))
                            (codep (cdr u))
                            (eqcar fnexp 'lambda)
                        )
                        (flag1 fnname 'planttrampoline)
                 )


                 (cond
		      ((codep fnexp) 
				     (cflush (inf fnexp) 10000 0)
				     (makefcode fnname fnexp)
		                     (remprop fnname '*lambdalink))
		      ((and (idp fnexp)
			    (not (funboundp fnexp)))
		       (return (putd fnname fntype (cdr (getd fnexp)))))
		      (*comp
		            (return (compd fnname fntype fnexp)))
                    
                      ((and (flagp fnname 'planttrampoline)
                            (eqcar fnexp 'lambda))
                        (planttrampoline fnname (length (cadr fnexp)))
                        (put fnname '!*lambdalink fnexp)
                       )

		      ((eqcar fnexp 'lambda)
                                            (put fnname '*lambdalink fnexp)
					    (makeflambdalink fnname))
		      (t
                        (return
                               (conterror 1105
				  "Ill-formed function expression in PutD"
					  (putd fnname fntype fnexp)
					  )
			       )
			)
		      )

                 (if
		    (neq fntype 'expr)
		    % THEN
		    (progn (put fnname 'type fntype)
			   (if (wgetv SYMGET (idinf fnname))
				(wputv (inf (wgetv SYMGET (idinf fnname)))
					 1 fntype)
				(wputv SYMGET (idinf fnname)
					(mkvect &symgetsize&))
				(wputv (inf (wgetv SYMGET (idinf fnname)))
					 1 fntype)))
		    % ELSE
		    (remprop fnname 'type)
                    (when (wgetv SYMGET (idinf fnname)) 
                           (wputv (inf (wgetv SYMGET (idinf fnname))) 1 NIL)))

                 (if
		    *usermode
		    (flag (list fnname) 'user)
		    (remflag (list fnname) 'user))

		 (when
		      printredefinedmessage
		      % THEN
		      (errorprintf "*** Function %r has been redefined"
				   fnname))
                 (return fnname)
		 )
	   )
	)
  )


(compiletime (setq !*lapopt nil))

(lap '((*entry trampoline expr 0)
       (lda (reg t3) 0) % id number for compiledcallinginterpreted
       (LDA (REG T2) (IDLOC compiledcallinginterpreted))
       (S8ADDQ (REG T2) (REG SYMFNCCENTER) (REG R27))
       (LDQ (REG R27) (DISPLACEMENT (REG R27) 0))
       (JMP (REG NULL) (DISPLACEMENT (REG R27) NIL))
       (WORDALIGN)))

(de planttrampoline(u p)
   % install an indirect call to compiledcallinginterpreted
   (let ((m (gtbps 4))
         (n (id2int u))
         q
         (a (inf (cdr (getd 'trampoline)))) )

          % argument count
        (putmem m p)
          % insert the id number in the 1st instruction
          % this instruction is in the lower 32 bits of the 1st word
          % because of little endian. Also, lapopt must be off - otherwise
          % we don't know where this instruction can be found.
        (setq q (wor (getmem a) n))
        (putmem (wplus2 m 8) q)
        (putmem (wplus2 m 16) (getmem (wplus2 a 8)))
        (putmem (wplus2 m 24) (getmem (wplus2 a 16)))
           % now plant it
        (setf (getmem (wplus2 symfnc (wtimes2 n addressingunitsperitem)))
              (wplus2 m addressingunitsperitem))
          ))
%(console-print-string " rrecursivechannelprin2 ")
%(unixputn (cdr (getd 'recursivechannelprin2)))
%(console-newline)
