%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:BACKTRACE.SL 
% Title:        Functions for displaying the return stack. 
% Author:       ??? 
% Created:      ??? 
% Modified:     22-Mar-84 09:23:19 (Brian Beach) 
% Package:      Kernel 
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 27-May-87 (Harold Carr & Leigh Stoller)
%  Added fluid declaration of in* and out*.
% 22-Mar-84 09:13:24 (Brian Beach)
%  Fixed constants.  Removed WFORS.
% 01-Dec-83 14:12:23 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(on fast-integers)

(compiletime
  (flag '(backtrace1 backtracerange) 'internalfunction))

(fluid '(ignoredinbacktrace* options* interpreterfunctions*))
(fluid '(in* out*))

(setq ignoredinbacktrace*
      (list 'eval 'apply 'fastapply 'codeapply 'codeevalapply 'catch
            'errorset 'evprogn 'toploop 'breakeval 'bindeval 'break 'main
            'stackcheck))

(setq interpreterfunctions* (list 'cond 'prog 'and 'or 'progn 'setq))

(global '(heapupperbound stacklowerbound))

(lap '((*entry get-stack-pointer expr 0)
       (*alloc 0)
       (*move (reg st) (reg 1))
       (*exit 0)))

(de interpbacktrace ()
  (prog (here x)
     (setq here (get-stack-pointer))
     (printf "Backtrace, including interpreter functions, from top of stack:%n")
     (return (backtracerange here stacklowerbound 1))))

(compiletime
 (put 'memo 'opencode '((ldw (displacement (reg sr0) (reg 1) 0) (reg 1)))))

(de backtrace ()
  (prog (here )
        (setq here (get-stack-pointer))
        (printf "Backtrace from top of stack:%n")
        (return (backtracerange here stacklowerbound 0))))

(de backtracerange (starting ending interpflag)
  (setq starting (get-stack-pointer))
  (prog (x i end retour)
    (setq ending  (wshift ending 5))
    (setq end starting)
    (while (wgreaterp end ending)
     (setq retour (wplus2 end -20))
     (setq starting (wplus2 end -36))
     (setq end (memo (wplus2 end -4)))
     (when (wlessp end ending) (setq end ending))
     (when (setq x (returnaddressp (memo retour)))
	(backtrace1 x interpflag))
     (for (from ii starting (wplus2 4 end) -4)
         (do (setq i ii)
             (cond ((weq (tag (memo i)) btr-tag)
                    (backtrace1 (mkid (inf (memo i))) interpflag))))))
    (return (terpri))))

(de verbosebacktrace ()
  (prog (here x i stkl)
    (setq here (get-stack-pointer))
    (setq stkl (wshift stacklowerbound 5))
    (for % start a little before here
     (from ii here stkl
           (wminus (wtimes2 addressingunitsperitem stackdirection)))
     (do (setq i ii)
         (cond ((and (eq (tag (memo i)) 29)
                     (wgreaterp (inf (memo i)) heapupperbound))
                (writechar (char tab))
                (channelwriteunknownitem out* (memo i))
                (terpri))
               ((weq (tag (memo i)) btr-tag)
                (printf "       %r%n" (mkid (inf (memo i)))))
               ((setq x (returnaddressp (memo i)))
                (printf "%p -> %p:%n"
                        (code-address-to-symbol (memo i) ) x))
               (t (printf "     %p%n" (memo i))))))
    (return (terpri))))

    

(de backtrace1 (item code)
  %
  % Code is 1 if Interpreter functions should be printed, 0 if not.
  %
  (unless (memq item ignoredinbacktrace*)
    (unless (and (equal code 0) (memq item interpreterfunctions*))
      (if (and (wlessp (idinf item) maxsymbols) 
               (stringp (symnam (idinf item))))
		(prin1 item))
      (writechar (char blank)))))

(off fast-integers)
