%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:DUMPLISP.SL
% Description:  Dump running lisp to an a.out format file.
% Author:       RAM, HP/FSD
% Created:      27-Feb-84
% Modified:     14-Jan-85 09:10:20 (Vicki O'Day)
% Package:      
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
% 31-Aug-88 (Julian Padget)
%  Changed second "fluid" call to "global" since there was no need of former.
% 01-Jun-87 (Leigh Stoller & Harold Carr)
%  Unbundled dumplisp.
% 19-Dec-86 (Leigh Stoller)
%  Unexec now moves the text/data boundary. See bps.c in the $pb directory
%   for a full explanation.
% 09-Dec-86 (Leigh Stoller)
%  Removed calls to expand_file_name and put them into c-code. Problems with
%   malloc. Commentted out seconf reclaim calls for copying-gc. This needs to
%   be looked at. Should be an easier way to switch.
% 24-Nov-86 (Leigh Stoller)
%  Added calls to expand_file_name so that ~ and $ variables can be used in
%   .sl scripts to build bare-psl.
% 28-Dec-84 (Vicki O'Day)
%  Decided not to use nextbps; we may want to use shared Nmode
%  executables, and this would be very dangerous with some bps
%  in text area.
% 17-Oct-84 12:18:31 (RAM)
%  Now use NEXTBPS and HEAPLAST.  This works now that setupbpsandheap does
%  an initial malloc and free for some scratch space for library calls.
% 27-Feb-84 16:11:36 (RAM)
%  Print error messages from here, not from the foreign function unexec.
%  Don't call unexec with NEXTBPS and HEAPLAST (use 0's).  There seem to be
%  bugs when these are used.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(fluid '(symbolfilename*))

(global '(nextbps heaplast heapupperbound oldheapupperbound stacklowerbound))

(de dumplisp (filename)
  (dumplispaux filename nextbps heaplast))

(de dumplispaux (filename data-start bss-start)
  (prog (savedstacklowerbound unexecresult savedunixargs)
        (unless (stringp filename)
          (stderror "Dumplisp requires a filename argument"))
        (reclaim)
	(if (not (boundp 'gcarraylowerbound))
	  (if (wgreaterp heapupperbound oldheapupperbound)
	    (reclaim)))
        (setq savedstacklowerbound stacklowerbound)
        (setq stacklowerbound 0)
	(setq savedunixargs unixargs*)    % Force each new system to get 
	(setq unixargs* nil)              % its own args
        % must be 0 for a new file
        (setq unexecresult
	      (unexec
	       (unixstring filename)
	       (if symbolfilename*
		 (unixstring symbolfilename*)
		 (stderror "Failure in Unexec: symbolfilename* is nil"))
	       data-start    
	       bss-start))
        (setq stacklowerbound savedstacklowerbound)
	(setq unixargs* savedunixargs)
        (cond ((or (lessp unexecresult 0) (greaterp unexecresult 10))
               (stderror "Unknown error in unexec."))
              ((not (eqn 0 unexecresult))
               (stderror (concat "Unexec: "
                          (indx
                               [""
                                "failed to create dump file."
                                "failed to open symbolfilename*."
                                "failed to access /dev/kmem."
                                "failed to read header from symbolfilename*."
                                "failed to write header to dump file."
                                "failed to write text segment to dump file."
                                "failed to write data segment to dump file."
                                "failed to copy symbol table to dump file."
                                "symbolfilename* is not an a.out format file."
                                "failed to close the dump file."
                                ] unexecresult)))))
        ))

%% End of File.



