%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PNK:EVAL-WHEN.SL 
% Title:        Funny business to make things happen at different times 
% Author:       Eric Benson 
% Created:      30 August 1981 
% Modified:     06-Apr-84 11:04:06 (Brian Beach) 
% Status:       Experimental 
% Mode:         Lisp 
% Package:      Kernel 
% Compiletime:  
% Runtime:      
%
% (c) Copyright 1983, Hewlett-Packard Company, all rights reserved.
% (c) Copyright 1981, University of Utah
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 06-Apr-84 11:03:56 (Brian Beach)
%  Removed function STARTUPTIME.
% 21-Mar-84 15:00:04 (Jim Ambras)
%   Killed copyd of startuptime
% 01-Dec-83 14:43:05 (Brian Beach)
%   Translated from Rlisp to Lisp.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
%
% Functions flagged IGNORE are evaluated immediately when invoked at the top
% level while compiling to a file.  Those flagged EVAL are evaled immediately
% and also passed to the file.  These functions are defined to make those
% actions more visible and mnemonic.
%
 
(dm commentoutcode (u)
  nil)
 
%. Comment out a single expression
(de compiletime (u)
  %. Evaluate at compile time only
  u)
 
% just return the already evaluated argument
(flag '(commentoutcode . (compiletime)) 'ignore)
 
% The functions above need only be present at compile time.  Those below must
 
% be present at both compile and load time to be effective.
(de bothtimes (u)
  %. Evaluate at compile and load time
  u)
 
(flag '(bothtimes) 'eval)
 
(de loadtime (u)
  %. Evaluate at load time only
  u)
 
(remflag '(loadtime) 'ignore)
 
% just to be sure it doesn't
(remflag '(loadtime) 'eval)
 
% happen until load time
 
