%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:           relink.sl
% Title:          adjust pointers in preparation of memory layout changes
% Author:         Herbert Melenk, ZIB Berlin
% Created:        12 Januar 1986 (in compacting-gc.sl)
% Status:         Experimental
% Mode:           Lisp
%
% (c) Copyright 1986  Konrad Zuse-Zentrum Berlin, for modifications
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
% Revisions:
% 10-Sep-1988  W. Neun , ZIB Berlin.
%   Made separate file.  Adapted to new heap layout (lower and upper heap)
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  WARNING:
%  this algorithm expects, that
%   - BSP is followed by HEAP immediately
%   - WARRAYs are allocated in the upper part of BPS
%   - WARRAY pointers are tagged by a WORD-TAG
%  entry point: %relink
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
(compiletime  (progn
  (load vector-instructions inum if-system)
  (load fast-vector)
  (fluid '(relinklowerbound relinkshift))
  (put 'relinklowerbound 'Treg 8#67)
  (put 'relinkshift      'Treg 8#66)
  (put 'relinklowerbound 'registername "T67")
  (put 'relinkshift      'registername "T66")
 
  (flag '(relinkregion relinkheap relinkitem !%relink-internal
          gcerror
          relinkheapupper prerelink postrelink)  'internalfunction )
 
  (ds pointertagp (x)  %  x is a tag, which should be in range 1 .. 14
      (wgeq (wor (wdifference 14 x) (wdifference x 1)) 0 ))
 
  (put 'insinf 'opencode
          '((*putfield (reg 2) (reg 1) (wconst infstartingbit)
                                       (wconst infbitlength))))
 
  (put 'insinf 'destroys '((reg 1) (reg 2)))
 
  (ds signedinf (s) (signedfield s infstartingbit infbitlength))
 
   %macros for generation of pre- and post reclaim
   % save all fluids that have a *known-to-gc prop
 
  (de collectregs-read (x)
    (prog (nn)
     (when (and (get x 'registername) (setq nn (get x '*known-to-gc)))
        (setq collectedregs*
         (cons `(iputv gcextra ,nn ,x) collectedregs*)))))
 
   (dm idtoreg-conversion-read ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-read))
       (cons 'progn collectedregs*) )
 
   (de collectregs-write (x)
    (prog (nn)
      (when (and (get x 'registername) (setq nn (get x '*known-to-gc)))
        (setq collectedregs*
         (cons `(setf ,x (igetv gcextra ,nn)) collectedregs*)))))
 
   (dm idtoreg-conversion-write ()
       (setq collectedregs* nil)
       (mapobl (function collectregs-write))
       (cons 'progn collectedregs*) )
)) %end compiletime
 
(de prerelink (gcextra)     % save critical from register to heap
    (idtoreg-conversion-read))
 
(de postrelink (gcextra)     % restore critical from heap to register
    (idtoreg-conversion-write))
 
(if_system Cray-2
 (lap '((*entry !%relink expr 2)  % must update all T Registers too
       (*alloc 62)               % because they may be used for high frame
       (*move 0 (Reg T126))        (*move 0 (Reg T127))
       (*move 0 (Reg T130))        (*move 0 (Reg T131))
       (*move 0 (Reg T132))        (*move 0 (Reg T133))
       (*move 0 (Reg T134))        (*move 0 (Reg T135))
       (*move 0 (Reg T136))        (*move 0 (Reg T137))
       (*move 0 (Reg T140))
       (*call !%relink-internal) % frame
       (*linke 62 clear-extraregs expr 0)))
 
 (lap '((*entry !%relink expr 2)  % must update all T Registers too
       (*alloc 33)               % because they may be used for high
       (*move 0 (Reg T71))       % frame
       (*move 0 (Reg T72))
       (*move 0 (Reg T73))
       (*move 0 (Reg T74))
       (*move 0 (Reg T75))
       (*move 0 (Reg T76))
       (*call !%relink-internal) % frame
       (*linke 33 clear-extraregs expr 0))))
 
(de !%relink-internal( delta warraystrt)
  % delta: shift quantity
  % warraystrt: start of warrayspace inside BPS
  % NEW: warraystrt must not be warraystart but occasionally it is
  %      heaptrapbound
  %
  (prog (b gcextra)
        (setq stackend st)
        (setq gcextra gcextrasave)
        (setq stackstart stacklowerbound)
        (prerelink gcextra)
        (setq relinklowerbound warraystrt) % copying into Treg fluids
        (setq relinkshift delta)
        (relinksymbols)
        (relinkregion stackstart stackend)
        (setq b bndstklowerbound)
        (while (progn (setq b (adjustbndstkptr b 1)) (wleq b bndstkptr))
                        (relinkitem b (getmem b)))
        (relinkheap)
        (relinkheapupper)
        (postrelink gcextra)))
 
(de relinksymbols ()
  (ifor (from i 0  maxsymbols 1)
        (do (prog (name)
             (setq name (symnam i))
             (when (stringp name)
                     (relinkitem (loc (symnam i)) name)
                     (relinkitem (loc (symval i)) (symval i))
                     (relinkitem (loc (symget i)) (symget i))
                     (relinkitem (loc (symprp i)) (symprp i)))))))
 
(de relinkregion (low high)
  (ifor (from ptr low high 1) (do (relinkitem ptr (getmem ptr)))))
 
(de relinkheap ()
  (prog (currentitem oldcurrentitem ccont shift infccont)
    (setq shift relinkshift)
    (setq oldcurrentitem (setq currentitem heaplowerbound))
    (setq ccont (getmem currentitem))
    (while (wlessp currentitem heaplast)
       (setq infccont (inf ccont))
       (case (tag ccont)
         ((btr-tag code-tag negint-tag posint-tag id-tag unbound-tag) NIL)
         ((string-tag bytes-tag fixnum-tag floatnum-tag bignum-tag
           words-tag halfwords-tag pair-tag vector-tag evector-tag)
               (when (and  (wgeq (inf ccont) relinklowerbound)
                           (wleq (inf ccont) heapupperbound))
                 (putmem currentitem
                            (insinf ccont (wplus2 shift infccont)))))
         ((hbytes-tag)
              (setq currentitem
                   (wplus2 currentitem (strpack (signedinf ccont)))))
         ((hwords-tag)
              (setq currentitem
                   (wplus2 currentitem (wrdpack (signedinf ccont)))))
         ((hvect-tag)
             (prog (tmp)
               (setq tmp (signedinf ccont)) % veclen
               (setq currentitem (wplus2 currentitem 1))
               (ifor (from i 0 tmp 1) (do
                     (prog (tmp2 tmp3 inftmp2)
                        (setq tmp2 (getmem currentitem))
                        (setq tmp3 (tag tmp2))
                        (setq inftmp2 (inf tmp2))
                        (when (and (pointertagp tmp3)
                                   (wgeq inftmp2 relinklowerbound)
                                   (wleq inftmp2 heapupperbound))
                                (putmem currentitem
                                     (insinf tmp2 (wplus2 shift inftmp2))))
                        (setq currentitem (wplus2 currentitem 1)))))
                (setq currentitem (wplus2 currentitem -1))))
          ((forward-tag)
               (setq currentitem
                    (wplus2 -1
                       (wplus2 currentitem (skiplength currentitem)))))
          (nil
            (progn (center-dump currentitem 600)
              (gcerror "Internal error in relink phase at %o" currentitem))))
 
        (setq currentitem (wplus2 currentitem 1))
        (setq ccont (getmem currentitem)))))
 
(de relinkheapupper ()
  (prog (currentitem ccont shift rangeend infccont)
      (setq shift relinkshift)
      (setq currentitem  (wplus2 heapupperbound -2))
      (setq rangeend (wplus2 heaptrapbound 10))
      (while (wgreaterp currentitem rangeend)
         (setq ccont (getmem currentitem))
         (when (and (pointertagp (tag ccont))
                    (wgeq (setq infccont (inf ccont)) relinklowerbound)
                    (wleq infccont heapupperbound))
             (putmem currentitem  (insinf ccont (wplus2 shift infccont))))
         (setq currentitem (wplus2 currentitem -1))
) )  )
 
(de relinkitem (ptr item)
  (prog (tg info)
        (setq tg (tag item))
        (unless (pointertagp tg) (return nil))
        (setq info (inf item))
        (when (or (wlessp info relinklowerbound)
                  (wgreaterp info heapupperbound)) (return nil))
        (putmem ptr (mkitem tg (wplus2 relinkshift info)))
     %  (setf (inf (getmem ptr)) (reloc info))
) )
 
(de gcerror (message p)
  (errorprintf message p)
  (while t (abort)))      % final exit
