%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXNK:TRAP.SL
% Description:  Signal handling for ibmxa
% Author:       Vicki O'Day, HP Labs/CRC
% Created:      27-Feb-84
% Modified:     2-Jan-85 13:13:16 (Vicki O'Day)
% Mode:         Lisp
% Package:
% Status:       Experimental (Do Not Distribute)
%
% (c) Copyright 1984, Hewlett-Packard Company, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 18-April 1991 (Winfried Neun) Initial version for IBM XA
% 18-Aug-86 (Leigh Stoller)
%  Reworked so that signals can be added more easily. Also added calls to
%  external functions which set up the trap, and reinit the trap after a
%  signal is received.
% 2-Jan-85 10:21:47 (Vicki O'Day)
%  Instead of just calling stderror with a message about the kind of trap,
%  we now find the name of the routine that caused the trap and report
%  that as well.  CODE-ADDRESS-TO-SYMBOL is used to figure out the name.
% 27-Dec-84 10:55:56 (Vicki O'Day)
%  Added sigunwind, to pop the signal handler's stack frame before calling
%  stderror.  (Many thanks to Jim Davis for deciphering the contents of this
%  frame.)
% 17-Dec-84 10:24:41 (Vicki O'Day)
%  Added lots more signals to catch, to try to prevent those occasional
%  dumps out of Nmode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

(compiletime (flag '(signal-done build-trap-message) 'internalfunction))

(fluid '(*bruch* ibmxa-sigst ibmxa-sigaddr ibmxa-sigcode ibmxa-signumber
         *dump ibmxa-sigpc ibmxa-dumparray*))

% this is the interface for toploop

(de initializeinterrupts () 
          % setting float interrupts first, as described in
          % ieee_handler (3M)

    (ieee_handler (strbase (strinf "set"))
                  (strbase (strinf "common"))
		  (symfnc (id2int 'signalhandler)))

	  % all signals are initiated expect for those in the following list
          % they are ignored

    (for (from i 1 31) (do
       (when (not (memq i '(1 15 18 19 20 21 22 28))) (sigsetup i)))))

% this installs the handler

(de sigsetup (signumber)
    (psl_sigset signumber (symfnc (id2int 'signalhaendler))))

% this is the 'handler'

(setq ibmxa-dumparray* (gtwarray 40))

(lap '((*entry signalhaendler expr 0)
%       We save the information which is in 'scratch' registers.
%
	 (*alloc 0)
	 (*move 128 (reg r9))
	 (*mkitem (reg r9) id-tag)
	 (*move (displacement (reg r1) 0) (reg 1))
	 (*move (displacement (reg r1) 4) (reg 2))
	 (*move (reg 1) ($fluid ibmxa-signumber))
         (*move (reg 2) ($fluid ibmxa-sigpc))
	 (*move (reg st) ($fluid ibmxa-sigst))
         (*linke 0 lisprpv expr 0)))

(setq *bruch* 0)

(de lisprpv()
   (prog (mess)
    (setq *bruch* 0)
    (smalldump)
    (setq mess (atsoc ibmxa-signumber '((1 . "Hangup")
                          (2 . "Interrupt")
                          (3 . "Quit")
                          (4 . "Illegal Instruction")
                          (5 . "Trace Trap")
                          (6 . "IOT Instruction")
                          (7 . "EMT Instruction")
                          (8 . "Floating Point Exception")
                          (10 . "Bus Error")
			  (11 . "Segmentation Violation")
                          (12 . "Bad Args to System Call")
                          (13 . "Write on Pipe With None to Read")
                          (14 . "Alarm Clock")
                          (15 . "Software termination signal")  )))
    (when (eq ibmxa-signumber 8)
          (cond ((eq ibmxa-sigcode 16#c8)
                         (setq mess '( 0 . "Floating Point divide by zero")))
                ((eq ibmxa-sigcode 16#d4)
                         (setq mess '( 0 . "Floating Point overflow")))
                ((eq ibmxa-sigcode 16#cc)
                         (setq mess '( 0 . "Floating Point underflow")))))
    (setq ibmxa-sigaddr 0 ) % for gc
    (sigrelse ibmxa-signumber (symfnc (id2int 'signalhaendler)))
    (setq mess (if mess (cdr mess)
                  (bldmsg " Unknown signal type %d"  ibmxa-signumber)))
    (build-trap-message mess ibmxa-sigpc)
    (return (stderror))))

(de smalldump()
 (prog (sta)
  (when *dump
    (console-print-string "*** UNEXPECTED INTERRUPT ***")
    (console-newline)
    (console-print-string "*** System signal number: 16#")
    (unixputn ibmxa-signumber)
    (console-print-string " , signal stack: 16#")
    (unixputn ibmxa-sigst)
    (console-newline)
    (console-print-string "*** signal pc: 16#")
    (unixputn ibmxa-sigpc)
    (console-newline))
    (setq sta ibmxa-sigst)
    (setq sta (wgetv sta 1))
    (setq sta (wgetv sta 1))
    %(setq sta (wgetv sta 1))
   (when *dump
    (console-print-string "Reg r0 ") (unixputn (wgetv sta 5))
    (console-print-string "Reg r1 ") (unixputn (wgetv sta 6))
    (console-print-string "Reg r2 ") (unixputn (wgetv sta 7))
    (console-print-string "Reg r3 ") (unixputn (wgetv sta 8))
    (console-newline)
    (console-print-string "Reg r4 ") (unixputn (wgetv sta 9))
    (console-print-string "Reg r5 ") (unixputn (wgetv sta 10))
    (console-print-string "Reg r6 ") (unixputn (wgetv sta 11))
    (console-print-string "Reg r7 ") (unixputn (wgetv sta 12))
    (console-newline)
    (console-print-string "Reg r8 ") (unixputn (wgetv sta 13))
    (console-print-string "Reg r9 ") (unixputn (wgetv sta 14))
    (console-print-string "Reg r10 ") (unixputn (wgetv sta 15))
    (console-print-string "Reg r11 ") (unixputn (wgetv sta 16))
    (console-newline)
    (console-print-string "Reg r12 ") (unixputn (wgetv sta 17))
    (console-print-string "Reg r13 ") (unixputn (wgetv sta 18))
    (console-print-string "Reg r14 ") (unixputn (wgetv sta 3))
    (console-print-string "Reg r15 ") (unixputn (wgetv sta 4))
    (console-newline))
    (setq ibmxa-sigpc (inf (wgetv sta 8))) %dont know anything better
))

(compiletime (flag '(x-code-address-to-symbol) 'internalfunction))

(de build-trap-message (trap-type trap-addr)
    (let (extra-info)
      (setf extra-info (bldmsg "%w%w" " in "
                                 (x-code-address-to-symbol (inf trap-addr))))
      (bldmsg "%w%w" trap-type extra-info)))

(fluid '(code-address* closest-address* closest-symbol*))

(de x-code-address-to-symbol (code-address*)
  (let ((closest-symbol* ()) (closest-address* 0) numb)
        (mapobl #'(lambda (symbol)
                 (when (fcodep symbol)
                       (let ((address (inf (getfcodepointer symbol))))
                            (when (and (ileq address code-address*)
                                       (igreaterp address closest-address*))
                                  (setq closest-address* address)
                                  (setq closest-symbol* symbol))))))
       closest-symbol*))

(de code-address-to-symbol (ad) (x-code-address-to-symbol ad))
