%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PXU:MULS.SL
% Description:  support for BIGNUM package with double INUM operations
%               lap source for MIPS processor
% Author:       H. Melenk, W. Neun
% Created:      25 January 1989
% Modified:
% Mode:         Lisp
% Package:      Utilities
% Status:       Experimental
%
% (c) Copyright 1989, Konrad-Zuse-Zentrum, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Mips version by W. Neun, ZIB Berlin
%
%  this file is needed compiletime only
%
% Double length multiply and divide for the 68020 processor.
% These are the open codes which embed the extended arithmetic 
% operations such that they are accessible from compiled LISP top level.
% They map the 2*30 bit operations needed by BIG to the 2*32 bit operations
% of the processor by shifting etc.
(fluid '(*second-value*))
 
(bothtimes
(put 'wdivide 'opencode
		'((div (reg 1) (reg 2))
                  (Mfhi (reg 3))
		  (mflo (reg 1))
		  (*move (reg 3) ($fluid *second-value*))))
)

(put 'wtimesdouble 'opencode
     % Called with two parameters of maximum bbase size
     % the return value are the low order bits of the product.
     % The high order bits are placed in fluid variable.
     '(% double length unsigned mutiply;
       (MULTU (reg 1) (reg 2))
       (Mfhi (reg 3))
       (mflo (reg 1))
       (*WSHIFT (reg 3) 2)
       (*MOVE (reg 1) (reg 2))
       (*WSHIFT (reg 2) -30)
       (*WOR (reg 3)(reg 2))
       (*MOVE (reg 3)($FLUID *second-value*))
       (*WSHIFT (reg 1) 2)
       (*WSHIFT (reg 1) -2) ))
 

(put 'wxtimes2 'opencode % different version for $pxu/mbarith
     '((MULT (reg 1) (reg 2))
      (mfhi  (reg 4))
      (mflo  (reg 1))
      (*move (reg 4) ($FLUID *second-value*))))


(commentoutcode
(de wquotientdouble (X Y Z)
      (PROG (A B C D)
               (SETQ A (oldWQUOTIENTDOUBLE X Y Z))
               (SETQ B *SECOND-VALUE*)
               (SETQ C (WQUOTIENTDOUBLELAP X Y Z))
               (SETQ D *SECOND-VALUE*)
               (unless (AND (EQ A C) (EQ B D)) (print (list "falsch"
                         x y z "," a b c d)) (quit))
               (return c)))


)
(fluid '(*second-value*))

(compiletime (progn
 
 (defcmacro *wrshift3)

 (de *wrshift3 (a1 a2 a3)
   (prog (resultingcode*)
     (setq a1 (resolveoperand '(reg t1) a1))
     (setq a2 (resolveoperand '(reg t1) a2))
     (setq a3 (resolveoperand '(reg t1) a3))
     (return `((*move ,a2 (reg t1))(srl ,a1 (reg t1) ,a3)))))
))

(commentoutcode % this is the algorithm for the lap hack behind:
(de wquotientdouble (arg1 arg2 arg3)

   (prog (quot rem carrybit i)

             % first divide the uper 30 bits by divisor and put the
             % quotient into the result and continue with the remainder
 
     (setq quot (wdivide arg1 arg3))
     (setq rem *second-value*)
 
     (setq arg2 (wshift arg2 2))  %initial shift
 
     (ifor (from i 0 29 1) (do
 
            % now shift the remainder left by 1 and add the most significant
            % bit of arg2. shift arg2 left one bit
            % if sum is greater or equal divisor, add a bit '1' to remainder
            % else add a '0'.
            % do it until arg2 is eaten up. quot is the result (quotient),
            %          rem is the remainder (in *second-value*)
 
            (progn
             (setq quot (wshift quot 1))
             (setq carrybit (wshift arg2 -31))
             (setq arg2 (wshift arg2 1))
             (setq rem (wplus2 (wshift rem 1) carrybit))
             (when (wgeq rem arg3) (setq rem (wdifference rem arg3))
                                   (setq quot (wplus2 quot 1)))
      )     )              )
     (setq *second-value* rem)
     (return quot)))
)
 
(lap '((*entry wquotientdouble expr 3)
       (*alloc 6)
       (*move (reg 1) (reg t1)) % arg1
       (*move (reg 2) (reg t2)) % arg2
       (*move (reg 3) (reg t3)) % arg3 
%      (*move (reg 3) (reg 2))
%      (*link wdivide expr 2)
       (*move (quote 0) (reg t4))  % quot
       (*move (reg 1) (reg t7)) % rem
       (*wshift (reg t2) 2)   
                                   30 times!
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab1) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab1
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab2) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab2
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab3) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab3
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab4) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab4
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab5) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab5
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab6) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab6
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab7) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab7
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab8) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab8
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab9) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab9
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab10) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab10
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab11) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab11
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab12) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab12
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab13) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab13
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab14) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab14
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab15) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab15
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab16) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab16
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab17) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab17
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab18) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab18
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab19) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab19
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab20) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab20
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab21) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab21
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab22) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab22
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab23) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab23
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab24) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab24
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab25) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab25
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab26) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab26
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab27) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab27
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab28) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab28
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab29) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab29
       (*wshift (reg t4) 1)
       (*wrshift3 (reg 2) (reg t2) 31) % carrybit
       (*wshift (reg t2) 1) 
       (*wshift (reg t7) 1) 
       (*wplus2 (reg t7) (reg 2))
       (*jumpwlessp (label lab30) (reg t7) (reg t3))
       (*wdifference (reg t7) (reg t3))
       (*wplus2 (reg t4) 1)
    lab30                           %done
       (*move (reg t7) ($fluid *second-value*))
       (*move (reg t4) (reg 1))
       (*exit 6)))


(commentoutcode de wquotientdoublesmall (arg1 arg2 arg3 norm)

   (prog (quot summa)

             % first divide the uper 30 bits by divisor and put the
             % quotient into the result and continue with the remainder

     (when (eq 0 arg1) (return (wdivide arg2 arg3)))

     (setq arg2 (wshift arg2 2))  %initial shift

     (setq summa 0 quot 0)
   lupe
     (when (wgeq (iplus2 norm summa) 30) (setq norm  (wdifference 30 summa)))
     (setq summa (iplus2 summa norm))

     (setq arg1 (wor (wshift arg1  norm) (wshift arg2 (wdifference norm 32))))
     (setq arg2 (wshift arg2 norm))
     (setq quot (wor (wshift quot norm) (wdivide arg1 arg3)))
     (setq arg1 *second-value*)
     (when (neq summa 30) (go lupe))

     (return quot)))

(lap '((*ENTRY wquotientdoublesmall EXPR 4)
	(*ALLOC 0)
	(*MOVE (REG 1) (reg ff1))
	(*MOVE (REG 2) (reg 5))
	(*MOVE (REG 3) (reg ff2))
	(*MOVE (Reg 4) (reg ff3))
	(*JUMPNOTEQ (LABEL G0005) (REG 1) (QUOTE 0))
	(*MOVE (REG 3) (REG 2))
	(*MOVE (reg 5) (REG 1))
	(*LINKE 0 WDIVIDE EXPR 2)
	(*LBL (LABEL G0005))
	(*MOVE (reg 5) (REG 1))
	(*WSHIFT (REG 1) (QUOTE 2))
	(*MOVE (REG 1) (reg 5))
	(*MOVE (QUOTE 0) (reg t1))
	(*MOVE (QUOTE 0) (reg ff4))
	(*LBL (LABEL G0004))
	(*WPLUS3 (REG 1) (reg ff3) (reg t1))
	(*JUMPWLESSP (LABEL G0009) (REG 1) (QUOTE 30))
	(*WDIFFERENCE3 (REG ff3) (WCONST 30) (reg t1))
	(*LBL (LABEL G0009))
	(*WPLUS3 (REG t1) (reg t1) (reg ff3))
%(*MOVE (reg ff1) (REG 2))
%(*WSHIFT (REG 2) (reg ff3))
        (SLLV (REG 2) (REG ff1) (REG ff3))
	(*WPLUS3 (REG 4) (WCONST -32) (reg ff3))
%(*MOVE (REG 3) (REG 4))
%(*MOVE (reg 5) (REG 3))
%(*WSHIFT (REG 3) (REG 4))
        (SUB (REG T6) (REG NULL) (REG 4))
        (SRLV (REG 3) (REG 5) (REG t6))
	(*WOR (REG 2) (REG 3))
	(*MOVE (REG 2) (reg 1))
%(*WSHIFT (REG 5) (reg ff3))
        (SLLV (REG 5) (REG 5) (REG ff3))
	(*MOVE (reg ff2) (REG 2))
%(*MOVE (reg ff1) (REG 1))
	(*LINK WDIVIDE EXPR 2)
%(*MOVE (reg ff4) (REG 3))
%%(*WSHIFT (REG 3) (reg ff3))
        (SLLV (REG 3) (REG ff4) (REG ff3))
	(*WOR (REG 3) (REG 1))
	(*MOVE (REG 3) (reg ff4))
	(*MOVE ($FLUID *SECOND-VALUE*) (reg ff1))
	(*JUMPnotEQ (LABEL G0004) (reg t1) (QUOTE 30))
	(*MOVE (REG 3) (REG 1))
	(*EXIT 0)))

(put 'addcarry 'opencode
       '((*MOVE ($FLUID carry*) (reg 2))
         (*WPLUS2 (reg 1)(reg 2))
         %(*MOVE (reg 1)(reg 3))
         %(*WSHIFT (reg 3) -30)
         (srl (reg 3) (reg 1) 30)
         (*MOVE (reg 3) ($FLUID carry*))
         (*WSHIFT (reg 1) 2)
         (*WSHIFT (reg 1) -2)))

(put 'subcarry 'opencode
       '((*MOVE ($FLUID carry*) (reg 2))
         (*WDIFFERENCE (reg 1)(reg 2))
         %(*MOVE (reg 1) (reg 2))
         %(*WSHIFT (reg 2) -31)
         (srl (reg 2) (reg 1)  31 )
         (*MOVE (reg 2) ($FLUID carry*))
         (*WSHIFT (reg 2) 30)
         (*WPLUS2 (reg 1) (reg 2))
       ))
         

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%   functions supporting word arithmetic with bbase*
%   bbase* is too large to be stored in a fluid cell;
%   so we perform operations with bbase* by shift and mask
%   instructions.
 
%   get the value for bbase* into a register
(put 'bbase** 'opencode
%     '((*MOVE 1 (reg 1))
%       (*WSHIFT (reg 1) 30)))
      '((lui (reg 1) 16#4000 )))
 
%   get the value for logicalbits* into a register     
(put 'logicalbits** 'opencode
      '((*MOVE -1 (reg 1))
        (*WSHIFT (reg 1) -2)))

%   calculate (remainder x bbase*)
(put 'remainder-bbase 'opencode
      '((*WSHIFT (reg 1) 2)
        (*WSHIFT (reg 1) -2)))

%   calculate (quotient x bbase*)
(put 'quotient-bbase 'opencode
      '((*WSHIFT (reg 1) -30)))

(put 'bbase** 'destroys '((reg 1))) 
(put 'logicalbits** 'destroys '((reg 1)))  
(put 'remainder-bbase 'destroys '((reg 1)))  
(put 'quotient-bbase 'destroys '((reg 1)))  


