%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% File:         pslcomp.sl
% Description:  Interpret program command string for source files to compile
% Author:       Leigh Stoller & Stan Shebs
% Created:      22-Nov-86
% Package:      
%
% (c) Copyright 1986, University of Utah
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Revisions
%
% 02-Dec-86 (Leigh Stoller)
%  Removed load of useful since it is now a kernel module named macros.sl.
%   Corrected spelling of compiler.
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Redefines PSLCOMP start-up routine to read and interpret program command
%  string as a list of source files to be compiled.

(imports '(get-command-args compiler string-search))

(fluid '(*usermode *redefmsg *WritingFASLFile))

(cond ((funboundp 'original-main)
       (copyd 'original-main 'main)))

(de main ()
  (let ((c-list (get-command-args))
	(*usermode nil)
	(*redefmsg nil))
    (compile-files c-list)
    (copyd 'main 'original-main))
  (original-main)
  )

(de compile-files (c-list)
  (cond ((null c-list)
	 (PrintF "Portable Standard Lisp Compiler%n")
	 (PrintF "Usage: PSLCOMP source-file ...%n"))
	(t
	 (for (in fn c-list)
	      (do (attempt-to-compile-file fn)))
         (exitlisp))))

(de attempt-to-compile-file (fn)
  (let* ((*break NIL)
	 (result (ErrSet (compile-file fn) T)))
    (cond ((FixP result)
	   (if *WritingFASLFile (faslend))
	   (printf "%n ***** Error during compilation of %w.%n" fn)))))


% Returns either T (successful), ABORT (aborted by user after error is
% signalled), or NIL (source file not found).

(de compile-file (fn)
  (exit (compile-file-aux fn))

%   (let ((result (compile-file-aux fn)))
%     (cond ((eq result 'ABORT) (exit 'ABORT))
% 	  ((eq result T) (exit T))
%	  ((null result) (printf "Source file %w not found%n" fn)))))

% This function attempts to compile the specified source file.
% Returns T if successful, ABORT if aborted by the
% user after an error is signalled, NIL if the source file is not found.
% The compilation is terminated even upon an abnormal exit.

(de compile-file-aux (source-fn)
  (cond ((filep source-fn)
	 (let ((result 'ABORT))
	   (unwind-protect (setq result (unprotected-compile-file source-fn))
	     % Here is the cleanup action:
	     (when (eq result 'ABORT)
	       (faslabort)
	       (printf "%f----- Compilation of %w aborted.%n%n" source-fn)))
	   result))
	(t
	 (printf "Source file %w not found%n" fn)
	 nil)))

(de unprotected-compile-file (source-fn)
  (let* ((*quiet_faslout T)
	 (file-type (get-file-extension source-fn))
	 (binary-fn (get-binary-name source-fn file-type)))
    (printf "%n----- Compiling %w to %w.b -----%n"
	    source-fn binary-fn)
    (faslout binary-fn)
    (if (eq 'ABORT (dskin-file 'COMPILE-FILE source-fn NIL #'faslend))
      'ABORT
      (progn
	(printf "%f----- Compilation of %w completed.%n%n" source-fn)
	T))))

(de get-binary-name (filename type)
  (cond ((equal type 'SL)
	 (subseq filename 0 (difference (size filename) 2)))
	(t
	 filename)))

(de get-file-extension (filename)
  (cond ((string-search-equal ".sl" filename)
	 'SL)))
	

