%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File:         PU:FAST-VECTOR.SL 
% Description:  IGetV, IGetS, etc. 
% Author:       
% Created:      
% Modified:     19 Oct 1984 0952-PDT (Brian Beach) 
% Package:      Utilities 
%
% (c) Copyright 1987, University of Utah, all rights reserved.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Revisions:
%
% 7-Jul-1988 (T. Yamamoto and C. Burdorf)
%    used if_system MC68020 for SPARC.
% 27-Sep-1986 (Leigh Stoller)
%  Changed 68K versions of fast functions to do tag stripping since the 68020
%  uses 32 bit addressing, not 24 like the 68000.
% 19 Oct 1984 0952-PDT (Brian Beach)
%  Fixed FAST-IPUTS macro for DEC20 version.  Was missing a parameter.
% 29-Mar-84 14:02:01 (Brian Beach)
%  Changed for new fast function model in compiler using FAST-FUNCTION property.
%  1 Mar 1984 0032-PST (Nancy Kendzierski)
%   Changed PDP10 so it only strips tags for strings (for extended 20).
% 19 Jan 1984 1431-PST (Brian Beach)
%   Added standard header.
% 29 Jan 1983 1606-PST, Cris Perdue
%  More careful code imported from FAST-VECTORS.SL and FAST-STRINGS.SL
%  Code also converted to LISP
%  <PSL.UTIL>FAST-VECTOR.RED.1, 18-Mar-82 21:26:35, Edit by GRISS
%  Fast Vector operations
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Probably the best thing to do about obtaining needed SysLisp
% features is to make sure the compiler has them at all times.

(compiletime (load if-system))

(de IGetV (v i)
  (cond ((not (Vectorp v)) (NonVectorError v 'IGetV))
	((not (FixP i)) (NonIntegerError i 'IGetV))
	(t (indx v i))
	))

(de IPutv (v i x)
  (cond ((not (Vectorp v)) (NonVectorError v 'IPutv))
	((not (FixP i)) (NonIntegerError i 'IPutv))
	(t (setindx v i x))
	))

(de ISizeV (v)
  (cond ((not (Vectorp v)) (NonVectorError v 'ISizeV))
	(t (size v))
	))

(Put 'IGetV 'Assign!-Op 'IPutV)

(de IGetS (s i)
  (cond ((not (StringP s)) (NonStringError s 'IGetS))
	((not (FixP i)) (NonIntegerError i 'IGetS))
	(t (indx s i))
	))

(de IPutS (s i c)
  (cond ((not (StringP s)) (NonStringError s 'IPutS))
	((not (FixP i)) (NonIntegerError i 'IPutS))
	((not (FixP c)) (NonCharacterError c 'IPutS))
	(t (setindx s i c))
	))

(de ISizeS (s)
  (cond ((not (StringP s)) (NonStringError s 'ISizeS))
	(t (size s))
	))

(Put 'IGetS 'Assign!-Op 'IPutS)

% Compiler declarations for fast functions.

(put 'igetv    'fast-function   'fast-igetv)
(put 'iputv    'fast-function   'fast-iputv)
(put 'igets    'fast-function   'fast-igets)
(put 'iputs    'fast-function   'fast-iputs)
(put 'isizev   'fast-function   'fast-isizev)
(put 'isizes   'fast-function   'fast-isizes)

(put 'igetv    'fast-flag   t)
(put 'iputv    'fast-flag   t)
(put 'igets    'fast-flag   t)
(put 'iputs    'fast-flag   t)
(put 'isizev   'fast-flag   t)
(put 'isizes   'fast-flag   t)

%
% Tags need to be stripped on IBM 370
  (ds fast-igetv (vector index) (vecitm (vecinf vector) index))
  (ds fast-iputv (vector index value) (putvecitm (vecinf vector)
                                                     index value))
  (ds fast-igets (string index) (strbyt (strinf string) index))
  (ds fast-iputs (string index value) (putstrbyt (strinf string)
                                                   index value))
  (ds fast-isizev (vector) (veclen (vecinf vector)))
  (ds fast-isizes (string) (strlen (strinf string)))))

% End of file.
